import numpy as np
from tqdm import tqdm
import torch
import random
from torch import nn
from torch.utils.data import Dataset, DataLoader
from typing import Optional, Tuple, List
from itertools import cycle


def _normalize(x, x_min, x_max):
    return x * (x_max - x_min) + x_min


def _init_patch_square(
    image_size, image_channels, patch_size_percent, data_min, data_max
):
    image_size = image_size**2
    noise_size = image_size * patch_size_percent
    noise_dim = int(noise_size**0.5)
    patch = np.random.rand(1, image_channels, noise_dim, noise_dim)
    patch = _normalize(patch, data_min, data_max)
    return patch


def _train_epoch(
    model,
    patch,
    train_dl,
    loss_function,
    optimizer,
    target_label,
    data_min,
    data_max,
    device,
):
    patch_size = patch.shape[-1]
    train_loss = []
    target = None
    for x, y in train_dl:
        # x, y = torch.tensor(x), torch.tensor(y)
        optimizer.zero_grad()
        if target is None:
            target = torch.tensor(
                np.full(y.shape[0], target_label),
                dtype=torch.long,
                device=device,
            )
        image_size = x.shape[-1]

        indx = np.random.randint(0, image_size - patch_size, size=y.shape[0])
        indy = np.random.randint(0, image_size - patch_size, size=y.shape[0])

        images = x.to(device)
        for i in range(y.shape[0]):
            images[
                i,
                :,
                indx[i] : indx[i] + patch_size,
                indy[i] : indy[i] + patch_size,
            ] = patch

        adv_out = model(images)

        loss = loss_function(adv_out, target[: y.shape[0]])
        loss.backward()
        optimizer.step()
        with torch.no_grad():
            patch.data = torch.clamp(patch.data, min=data_min, max=data_max)
        train_loss.append(loss.item())
    epoch_loss = np.array(train_loss).mean()
    return epoch_loss


def _validate(model, patch, data_loader, loss_function, target_label, device):
    patch_size = patch.shape[-1]
    val_loss = []
    preds = []
    with torch.no_grad():
        for x, y in data_loader:
            y = torch.tensor(
                np.full(y.shape[0], target_label), dtype=torch.long
            ).to(device)
            image_size = x.shape[-1]

            indx = random.randint(0, image_size - patch_size)
            indy = random.randint(0, image_size - patch_size)

            images = x.to(device)
            images[
                :, :, indx : indx + patch_size, indy : indy + patch_size
            ] = patch
            adv_out = model(images)
            loss = loss_function(adv_out, y)

            val_loss.append(loss.item())
            preds.append(adv_out.argmax(axis=1).detach().cpu().numpy())
        val_loss = np.array(val_loss).mean()
        preds = np.concatenate(preds)
        percent_successful = (
            np.count_nonzero(preds == target_label) / preds.shape[0]
        )
        return val_loss, percent_successful


def _make_patch(
    dataset,
    batch_size,
    model,
    target_label,
    device,
    patch_percent=0.1,
    epochs=5,
    data_min=None,
    data_max=None,
    lr=0.05,
):
    print(f"Training patch for label {target_label}...")
    dataloader = DataLoader(dataset, batch_size=batch_size, num_workers=4,
                            pin_memory=True)
    # patch values will be clipped between data_min and data_max
    # so that patch will be valid image data.
    if data_max is None or data_min is None:
        for x, _ in tqdm(dataloader):
            if data_max is None:
                data_max = x.max().item()
            if data_min is None:
                data_min = x.min().item()
            if x.min() < data_min:
                data_min = x.min().item()
            if x.max() > data_max:
                data_max = x.max().item()

    model.to(device)
    for param in model.parameters():
        param.requires_grad = False
    model.eval()

    x, _ = next(iter(dataloader))
    sample_shape = x.shape

    patch = _init_patch_square(
        sample_shape[-1], sample_shape[1], patch_percent, data_min, data_max
    )
    patch = torch.tensor(patch, requires_grad=True, device=device)
    optim = torch.optim.Adam([patch], lr=lr, weight_decay=0.0)

    loss = torch.nn.CrossEntropyLoss()
    min_loss = None
    best_patch = None

    for epoch in range(epochs):
        epoch_loss = _train_epoch(
            model,
            patch,
            dataloader,
            loss,
            optim,
            target_label=target_label,
            data_min=data_min,
            data_max=data_max,
            device=device,
        )
        print(f"Patch {target_label} epoch {epoch} loss: {epoch_loss}")
        if min_loss is None or epoch_loss < min_loss:
            min_loss = epoch_loss
            best_patch = patch.cpu()

    val_loss, percent_successful = _validate(
        model, patch, dataloader, loss, target_label, device
    )
    return best_patch, val_loss, percent_successful

def train_adversarial_patches(
        model: nn.Module,
        dataset: Dataset,
        num_patches: int,
        batch_size: int,
        path: Optional[str],
        labels: Optional[Tuple[int]] = None,
        device: Optional[torch.device] = None,
) -> List[torch.Tensor] | None:
    """Train adversarial patches for a given model and dataset.
    If `path` is not `None`, the patches are saved to disk.
    Otherwise, they are returned as a list.

    Parameters
    ----------
    model : nn.Module
        The model for which the patches should be trained.
    dataset : Dataset
        Torch Dataset to use for training the patches.
    num_patches : int
        The number of patches to train.
    batch_size : int
        The batch size to use for training.
    path : Optional[str]
        The path to which the patches should be saved.
        If `None`, the patches are returned as a list.
        Default: `None`.
    labels : Optional[Tuple[int]], optional
        Tuple of labels to use for the patches.
        If `None`, the labels are assumed to be `range(num_patches)`.
        Default: `None`.
    device : Optional[torch.device], optional
        Device to use, by default None.

    Returns
    -------
    List[torch.Tensor] | None
        If `path` is `None`, a list of patches.
        Otherwise, `None`.
    """    
    if device is None:
        device = torch.device("cpu")
    model.to(device)
    model.eval()

    patch_labels = cycle(labels) if labels is not None else range(num_patches)
    all_patches = []
    for patch_label in patch_labels:
        patch, val_loss, percent_successful = _make_patch(
            dataset, batch_size, model, patch_label, device
        )
        print(
            f"Patch label: {patch_label}.",
            f"Loss: {val_loss:.3f}.",
            f"Success rate: {percent_successful:.3f}.",
        )
        if path is not None:
            torch.save(patch, path + f"_{patch_label}.pt")
        else:
            all_patches.append(patch)
    if path is None:
        return all_patches


from attribench.data import IndexDataset
from itertools import cycle
import os
from ..._message import PartialResultMessage
from .._metric_worker import MetricWorker, WorkerConfig
from typing import Callable
from torch import nn

from attribench.result._grouped_batch_result import GroupedBatchResult
from attribench._method_factory import MethodFactory
from attribench.functional.metrics._impact_coverage import (
    _impact_coverage_batch,
)


class ImpactCoverageWorker(MetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: Callable[[], nn.Module],
        dataset: IndexDataset,
        batch_size: int,
        method_factory: MethodFactory,
        patch_folder: str,
    ):
        super().__init__(
            worker_config,
            model_factory,
            dataset,
            batch_size,
        )
        self.patch_folder = patch_folder
        self.method_factory = method_factory
        patch_names = [
            filename
            for filename in os.listdir(self.patch_folder)
            if filename.endswith(".pt")
        ]
        self.patch_names_cycle = cycle(patch_names)

    def setup(self):
        self.model = self._get_model()
        self.method_dict = self.method_factory(self.model)

    def work(self):
        self.setup()

        for batch_indices, batch_x, batch_y in self.dataloader:
            # Compute batch result
            batch_result = _impact_coverage_batch(
                self.model,
                self.method_dict,
                batch_x,
                batch_y,
                self.patch_folder,
                self.patch_names_cycle,
                self.device,
            )
            # Return batch result
            self.worker_config.send_result(
                PartialResultMessage(
                    self.worker_config.rank,
                    GroupedBatchResult(batch_indices, batch_result),
                )
            )



from attribench.masking.image import ImageMasker
import torch


class SampleAverageImageMasker(ImageMasker):
    """Image masker that masks pixels or features by replacing them with
    the average value in the corresponding image.
    """
    def __init__(self, feature_level: str):
        """
        Parameters
        ----------
        feature_level : str
            The level at which to mask the image. Must be either
            ``"pixel"`` or ``"feature"``.
        """
        super().__init__(feature_level)

    def _initialize_baselines(self, samples: torch.Tensor):
        batch_size, num_channels, rows, cols = samples.shape
        self.baseline = (
            torch.mean(samples.flatten(2), dim=-1)
            .reshape(batch_size, num_channels, 1, 1)
            .repeat(1, 1, rows, cols)
        )



from ..deletion._deletion import Deletion
from attribench._model_factory import ModelFactory
from attribench.masking import Masker
from attribench.data import AttributionsDataset
from attribench.result._insertion_result import InsertionResult
from typing import Dict, List, Optional, Tuple, Union


class Insertion(Deletion):
    """Compute the Insertion metric for a given :class:`~attribench.data.AttributionsDataset` and model
    using multiple processes. Insertion can be viewed as an opposite version
    of the Deletion metric.

    Insertion is computed by iteratively revealing the top (Most Relevant First,
    or MoRF) or bottom (Least Relevant First, or LeRF) features of
    the input samples, leaving the other features masked out,
    and computing the confidence of the model on the masked samples.

    This results in a curve of confidence vs. number of features masked. The
    area under (or equivalently over) this curve is the Insertion metric.

    `start`, `stop`, and `num_steps` are used to determine the range of features
    to mask. The range is determined by `start` and `stop` as a percentage of
    the total number of features. `num_steps` is the number of steps to take
    between `start` and `stop`.

    The Insertion metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`. The number of processes is
    determined by the number of devices. If `devices` is None, then all
    available devices are used. Samples are distributed evenly across the
    processes.

    Note that the Insertion metric is equivalent to the Deletion metric
    with the following changes:
    - Start and stop are 1 - start and 1 - stop, respectively
    - The mode parameter is swapped

    Note also that, if start and stop are 1 and 0 or vice versa, then
    Insertion-morf and Deletion-lerf are equal, and Insertion-lerf
    and Deletion-morf are equal.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        maskers: Dict[str, Masker],
        activation_fns: Union[List[str], str],
        mode: str = "morf",
        start: float = 0.0,
        stop: float = 1.0,
        num_steps: int = 100,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        attributions_dataset : AttributionsDataset
            Dataset containing the samples and attributions to compute
            Insertion on.
        batch_size : int
            The batch size to use when computing the metric.
        maskers : Dict[str, Masker]
            Dictionary of maskers to use for computing the metric.
        activation_fns : Union[List[str], str]
            List of activation functions to use for computing the metric.
            If a single string is given, it is converted to a single-element
            list.
        mode : str, optional
            Mode to use for computing the metric. Either "morf" or "lerf".
            Default: "morf"
        start : float, optional
            Relative start of the range of features to mask. Must be between 0 and 1.
            Default: 0.0
        stop : float, optional
            Relative end of the range of features to mask. Must be between 0 and 1.
            Default: 1.0
        num_steps : int, optional
            Number of steps to use for the range of features to mask.
            Default: 100
        address : str, optional
            Address to use for the multiprocessing connection.
            Default: "localhost"
        port : str, optional
            Port to use for the multiprocessing connection.
            Default: "12355"
        devices : Optional[Tuple], optional
            Devices to use. If None, then all available devices are used.
            Default: None
        """
        super().__init__(
            model_factory,
            attributions_dataset,
            batch_size,
            maskers,
            activation_fns,
            "lerf" if mode == "morf" else "morf",  # Swap mode
            1 - start,  # Swap start
            1 - stop,  # Swap stop
            num_steps,
            address,
            port,
            devices,
        )
        self._result = InsertionResult(
            attributions_dataset.method_names,
            list(maskers.keys()),
            list(self.activation_fns),
            mode,
            num_samples=attributions_dataset.num_samples,
            num_steps=num_steps,
        )



from typing_extensions import override
from numpy import typing as npt
from typing import List, Tuple, Optional
from ._deletion_result import DeletionResult
import pandas as pd


class InsertionResult(DeletionResult):
    """
    This class serves as a simple wrapper class for Insertion results.
    It is nearly identical to the DeletionResult class, except that
    the ``higher_is_better`` flag in ``get_df`` is inverted.
    """

    def get_df(
        self,
        masker: str,
        activation_fn: str,
        agg_fn="auc",
        methods: Optional[List[str]] = None,
        columns: Optional[npt.NDArray] = None,
    ) -> Tuple[pd.DataFrame, bool]:
        df, higher_is_better = super().get_df(
            masker, activation_fn, agg_fn, methods, columns
        )
        # The only difference between Deletion and Insertion is the fact
        # that the higher_is_better flag is inverted.
        return df, not higher_is_better

    @classmethod
    @override
    def _load(cls, path: str, format="hdf5") -> "InsertionResult":
        tree, mode = cls._load_tree_mode(path, format)
        res = InsertionResult(
            tree.levels["method"],
            tree.levels["masker"],
            tree.levels["activation_fn"],
            mode,
            tree.shape[0],
            tree.shape[1],
        )
        res.tree = tree
        return res


from typing import Dict, Tuple
import pandas as pd
import matplotlib.pyplot as plt

def mad_ratio(df):
    # Basically the F-ratio but using MAD instead of SS
    # MAD = median of absolute deviation to median
    # Mean of MADs for each method
    group_medians = df.median()
    within_mad = df.sub(group_medians).abs().median().mean()

    # MAD of group medians to global median
    global_median = df.stack().median()
    #between_mad = group_medians.sub(global_median).abs().median()
    between_mad = df.stack().sub(global_median).abs().median()

    return between_mad / within_mad


class MADRatioPlot:
    """
    Bar plot of the MAD ratio for each metric.
    The MAD ratio is the ratio of the total MAD to the MAD within groups.
    If this ratio is high, it means that there are differences in method
    behaviour.
    If the ratio is close to 1, it means that the methods behave similarly.
    This can be viewed as a one-way ANOVA test, but using non-parametric MAD
    instead of SS.
    """
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs

    def render(self, title=None, fontsize=20, figsize=(10, 10)):
        ratios = {}
        for metric_name, (df, inverted) in self.dfs.items():
            ratios[metric_name] = mad_ratio(df)
        df = pd.DataFrame(ratios, index=["MAD Ratio"]).transpose()
        fig, ax = plt.subplots()
        plt.xticks(fontsize=fontsize)
        plt.yticks(fontsize=fontsize)
        df.plot.barh(figsize=figsize, ax=ax)
        ax.set_title(title)
        fig.tight_layout()
        return fig



from abc import abstractmethod
from ._worker import (
    Worker,
    WorkerConfig,
    DistributedWorkerConfig,
    SynchronizedWorkerConfig,
)
from ._message import PartialResultMessage, DoneMessage
import torch.multiprocessing as mp
from typing import Optional, Tuple
import torch
import os


class DistributedComputation:
    def __init__(
        self,
        address: str,
        port: str | int,
        devices: Optional[Tuple[int]] = None,
    ):
        self.address = address
        self.port = str(port)
        self.devices = (
            devices
            if devices is not None
            else list(range(torch.cuda.device_count()))
        )
        self.world_size = len(self.devices)
        self.ctx = mp.get_context("spawn")

    @abstractmethod
    def _handle_result(self, result: PartialResultMessage):
        raise NotImplementedError

    @abstractmethod
    def _create_worker(self, worker_config: WorkerConfig) -> Worker:
        raise NotImplementedError

    def _cleanup(self):
        pass

    def run(self, *args, **kwargs):
        if self.world_size != 1:
            # Initialize multiproc parameters
            os.environ["MASTER_ADDR"] = self.address
            os.environ["MASTER_PORT"] = self.port

            result_queue = self.ctx.Queue()  # Will store results from metric
            # Used for signaling processes that they can terminate
            all_processes_done = self.ctx.Event()
            processes = []  # Used for joining all processes

            # Start all processes
            for rank in range(self.world_size):
                worker_config = DistributedWorkerConfig(
                    self.world_size, result_queue, rank, all_processes_done
                )
                worker = self._create_worker(worker_config)
                p = self.ctx.Process(target=worker.run)
                p.start()
                processes.append(p)

            # Gather results
            # This is not busy waiting:
            # queue.get will block until a result is passed
            done_processes = [False for _ in processes]
            while not all(done_processes):
                # res is a PartialResultMessage or a DoneMessage
                res = result_queue.get()
                if isinstance(res, DoneMessage):
                    done_processes[res.rank] = True
                else:
                    self._handle_result(res)

            # Processes are now allowed to terminate
            all_processes_done.set()
            for p in processes:
                p.join()
            self._cleanup()
        else:
            # If world size is 1, run on the main process
            worker_config = SynchronizedWorkerConfig(self._handle_result)
            worker = self._create_worker(worker_config)
            worker.run()
            self._cleanup()



import pandas as pd
from scipy import stats
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.figure import Figure
from attribench.plot import Plot


class CLESPlot(Plot):
    """Bar plot showing the CLES (Common Language Effect Size) of two methods.
    The CLES is the probability that the metric value for a random sample from
    one method will be greater than a random sample from the other method.

    This plot shows the CLES as a bar vertical bar plot centered around 0.5
    (no difference between methods). If the bar is on the right side of 0.5,
    it means that the first method is better than the second method. A bar
    is shown for each metric.
    """

    def render(self, method1: str, method2: str) -> Figure:
        """Render the plot.
        TODO add options for figsize, fontsize, title

        Parameters
        ----------
        method1 : str
            First method to compare.
        method2 : str
            Second method to compare.

        Returns
        -------
        Figure
            Rendered Matplotlib figure.

        Raises
        ------
        ValueError
            If one of the methods is not found in the dataframe.
        """
        result_cles = {}
        for key in self.dfs:
            df, higher_is_better = self.dfs[key]
            if not higher_is_better:
                df = -df

            if method1 not in df.columns:
                raise ValueError(
                    f"Method {method1} not found in dataframe {key}"
                )
            if method2 not in df.columns:
                raise ValueError(
                    f"Method {method2} not found in dataframe {key}"
                )

            statistic, pvalue = stats.wilcoxon(
                x=df[method1], y=df[method2], alternative="two-sided"
            )
            cles = (df[method1] > df[method2]).mean()
            result_cles[key] = cles if pvalue < 0.01 else 0.5
        sns.set_color_codes("muted")
        df = (
            pd.DataFrame(result_cles, index=["CLES"])
            .transpose()
            .reset_index()
            .rename(columns={"index": "Metric"})
        )
        df["CLES"] -= 0.5
        fig, ax = plt.subplots(figsize=(5, 7))
        sns.barplot(data=df, x="CLES", y="Metric", color="b", left=0.5, ax=ax)
        ax.set(xlim=(0, 1))
        ax.set_title(f"P({method1} > {method2})")
        return fig



from typing import Tuple, Dict, Optional
from ..._worker import WorkerConfig
from attribench.masking import Masker
from ._minimal_subset_worker import MinimalSubsetWorker
from attribench.result import MinimalSubsetResult
from .._metric import Metric
from attribench.data import AttributionsDataset
from attribench._model_factory import ModelFactory


class MinimalSubset(Metric):
    """Computes Minimal Subset Deletion or Insertion for a given
    :class:`~attribench.data.AttributionsDataset` and model using multiple processes.

    Minimal Subset Deletion or Insertion is computed by iteratively masking
    (Deletion) or revealing (Insertion) the top features of the input samples
    and computing the prediction of the model on the masked samples.

    Minimal Subset Deletion is the minimal number of features that must be
    masked to change the model's prediction from its original prediction.
    Minimal Subset Insertion is the minimal number of features that must be
    revealed to get the model's original prediction.

    The Minimal Subset metric is computed for each masker in `maskers`.
    The number of processes is determined by the number of devices. If `devices`
    is None, then all available devices are used. Samples are distributed evenly
    across the processes.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        maskers: Dict[str, Masker],
        mode: str = "deletion",
        num_steps: int = 100,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        attributions_dataset : AttributionsDataset
            Dataset containing the samples and attributions to compute
            the Minimal Subset metric for.
        batch_size : int
            Batch size per subprocess to use when computing the metric.
        maskers : Dict[str, Masker]
            Dictionary mapping masker names to `Masker` objects.
        mode : str, optional
            "deletion" or "insertion", by default "deletion"
        num_steps : int, optional
            Number of steps to use when computing the Minimal Subset metric,
            by default 100. More steps will result in a more accurate metric,
            but will take longer to compute.
        address : str, optional
            Address to use for multiprocessing, by default "localhost"
        port : str, optional
            Port to use for multiprocessing, by default "12355"
        devices : Optional[Tuple], optional
            Tuple of devices to use for multiprocessing, by default None.
            If None, all available devices are used.

        Raises
        ------
        ValueError
            If `mode` is not "deletion" or "insertion".
        """
        super().__init__(
            model_factory, attributions_dataset, batch_size, address, port, devices
        )
        self.dataset = attributions_dataset
        self.num_steps = num_steps
        if mode not in ["deletion", "insertion"]:
            raise ValueError("Mode must be deletion or insertion. Got:", mode)
        self.mode = mode
        self.maskers = maskers
        self._result = MinimalSubsetResult(
            attributions_dataset.method_names,
            list(maskers.keys()),
            mode,
            num_samples=attributions_dataset.num_samples,
        )

    def _create_worker(
        self, worker_config: WorkerConfig
    ) -> MinimalSubsetWorker:
        return MinimalSubsetWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.maskers,
            self.mode,
            self.num_steps,
        )



from abc import abstractmethod
from ._grouped_batch_result import GroupedBatchResult
from ._metric_result import MetricResult


class GroupedMetricResult(MetricResult):
    """
    Abstract class to represent grouped results of distributed metrics.
    These are results of metrics where the attributions dataset is grouped,
    i.e. the metric is computed for all attribution methods at a time.
    This is used for metrics  which have a shared computation
    for all attribution methods to save computation time, e.g. Infidelity.
    """

    @abstractmethod
    def add(self, batch_result: GroupedBatchResult):
        """
        Add a batch to the result (abstract method).
        """
        self.tree.write_dict(
            batch_result.indices.detach().cpu().numpy(),
            batch_result.results,
            level_order=self.level_order,
        )



from ._metric import Metric
from .deletion._deletion import Deletion
from .insertion._insertion import Insertion
from .impact_coverage._impact_coverage import ImpactCoverage
from .irof._irof import Irof
from .infidelity._infidelity import Infidelity
from .max_sensitivity._max_sensitivity import MaxSensitivity
from .minimal_subset._minimal_subset import MinimalSubset
from .sensitivity_n._sensitivity_n import SensitivityN
from .parameter_randomization._parameter_randomization import (
    ParameterRandomization,
)



from typing import Dict, Tuple
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.preprocessing import MinMaxScaler
import numpy as np
from scipy.cluster.hierarchy import linkage


class ClusterPlot:
    """
    Clustermap of the median values of the metrics.
    Allows the user to see which metrics and/or methods behave similarly.
    """
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs

    def render(self, figsize=(7, 7)):
        medians = {}
        for metric_name, (df, higher_is_better) in self.dfs.items():
            medians[metric_name] = (
                df.median(axis=0) if higher_is_better else -df.median(axis=0)
            )

        df = pd.DataFrame.from_dict(medians)
        normalized = MinMaxScaler().fit_transform(df)
        normalized_df = pd.DataFrame(
            normalized, columns=df.columns, index=df.index
        )

        # Manually computing the linkage so that we can set 
        # optimal_ordering to True
        # This should allow for easy comparison between clustermaps
        row_linkage = linkage(
            normalized,
            method="single",
            metric="correlation",
            optimal_ordering=True,
        )
        col_linkage = linkage(
            np.transpose(normalized),
            method="single",
            metric="correlation",
            optimal_ordering=True,
        )
        fig = sns.clustermap(
            normalized_df,
            row_linkage=row_linkage,
            col_linkage=col_linkage,
            figsize=figsize,
        ).fig
        return fig



from attribench.masking.image import ImageMasker
import torch
import numpy as np
from typing import List, Union


# TODO this should not be a subclass of ImageMasker
class TabularMasker(ImageMasker):
    def __init__(self, mask_value: Union[float, List[float]] = 0.0):
        self.mask_value = mask_value
        super().__init__("channel")

    def _initialize_baselines(self, samples):
        mask = torch.tensor(
            self.mask_value, device=samples.device, dtype=samples.dtype
        )
        self.baseline = (
            torch.ones(
                samples.shape, device=samples.device, dtype=samples.dtype
            )
            * mask
        )

    def _check_attribution_shape(self, samples, attributions):
        check1 = super()._check_attribution_shape(samples, attributions)
        if not isinstance(self.mask_value, float):
            return check1 and len(self.mask_value) == attributions.shape[-1]
        else:
            return check1

    def _mask(self, indices: np.ndarray):
        if self.baseline is None:
            raise ValueError("Masker was not initialized.")
        batch_size = self.samples.shape[0]
        num_indices = indices.shape[1]
        batch_dim = np.tile(range(batch_size), (num_indices, 1)).transpose()

        # to_mask = torch.zeros(samples.shape).flatten(1 if self.feature_level == "channel" else 2)
        to_mask = np.zeros(self.samples.shape)
        if self.masking_level == "channel":
            to_mask = to_mask.reshape((to_mask.shape[0], -1))
        else:
            to_mask = to_mask.reshape((to_mask.shape[0], to_mask.shape[1], -1))
        if self.masking_level == "channel":
            to_mask[batch_dim, indices] = 1.0
        else:
            try:
                to_mask[batch_dim, :, indices] = 1.0
            except IndexError:
                raise ValueError(
                    "Masking index was out of bounds. "
                    "Make sure the masking policy is compatible with method output."
                )
        to_mask = to_mask.reshape(self.samples.shape)
        return self._mask_boolean(to_mask)



from typing import Dict, Tuple
from numpy import typing as npt
import pandas as pd
from attribench._stat import wilcoxon_tests
from attribench.plot import Plot
from matplotlib.figure import Figure
from matplotlib import pyplot as plt
import numpy as np


def _effect_size_barplot(
    effect_sizes, pvalues, labels, alpha, figsize, fontsize
) -> Tuple[Figure, npt.NDArray]:
    fig, axs = plt.subplots(ncols=2, gridspec_kw={"width_ratios": [4, 2]})

    effect_sizes.plot.barh(figsize=figsize, ax=axs[0])
    axs[0].legend(
        labels,
        loc="upper center",
        bbox_to_anchor=(0.5, -0.05),
        ncol=3,
        fancybox=True,
        shadow=True,
    )
    axs[1].pcolor(
        pvalues < alpha, cmap="RdYlGn", edgecolor="black", vmin=0.0, vmax=1.0
    )
    axs[1].set_title(f"p < {alpha}")
    axs[1].set_yticks([])
    axs[1].set_xticks(np.arange(len(labels)) + 0.5, fontsize=fontsize)

    axs[1].set_xticklabels(
        labels,
        ha="right",
        rotation=45,
        rotation_mode="anchor",
        fontsize=fontsize,
    )
    return fig, axs


class WilcoxonBarPlot(Plot):
    """
    Alternative to WilcoxonSummaryPlot.
    Shows a bar plot of effect sizes, along with a grid showing statistical
    significance (green if significant, red otherwise).
    Provides more detailed information about effect size than the summary plot.
    Can get crowded if there are many methods/metrics.
    """
    def render(
        self,
        title: str | None = None,
        figsize=(12, 6),
        fontsize: int | None = None,
    ) -> Figure:
        """Render the plot.

        Parameters
        ----------
        title : str | None, optional
            Title of the figure, by default None
        figsize : tuple, optional
            Size of the figure, by default (12, 6)
        fontsize : int | None, optional
            Font size of x and y axis ticks, by default None

        Returns
        -------
        Figure
            Rendered Matplotlib figure.
        """
        ALPHA = 0.01
        effect_sizes, pvalues = {}, {}

        for metric_name, (df, higher_is_better) in self.dfs.items():
            # Compute effect sizes and p-values
            mes, mpv = wilcoxon_tests(df, higher_is_better)
            effect_sizes[metric_name] = mes
            pvalues[metric_name] = mpv

        es_df = pd.DataFrame.from_dict(effect_sizes)
        pv_df = pd.DataFrame.from_dict(pvalues)
        fig, axs = _effect_size_barplot(
            es_df,
            pv_df,
            self.dfs.keys(),
            ALPHA,
            figsize=figsize,
            fontsize=fontsize,
        )
        if title is not None:
            fig.suptitle(title)
        return fig



from abc import abstractmethod
from typing import Tuple, Dict
import pandas as pd
from matplotlib.figure import Figure


class Plot:
    """Abstract base class for all plots.
    A plot is simply an object that has a render method that returns a 
    matplotlib Figure.
    """
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        """
        Parameters
        ----------
        dfs : Dict[str, Tuple[pd.DataFrame, bool]]
            A dictionary mapping metric names to tuples of dataframes and
            booleans. The boolean indicates whether higher values of the metric
            are better (``True``) or not (``False``). The dataframes should
            have the same columns, which are the names of the methods.
        """
        self.dfs = dfs

    @abstractmethod
    def render(self, *args, **kwargs) -> Figure:
        raise NotImplementedError



from attribench.masking.image import ImageMasker
import torch


class RandomImageMasker(ImageMasker):
    """Image masker that masks images with normally distributed random
    noise, with a given standard deviation.
    """
    def __init__(self, masking_level: str, std=1):
        """
        Parameters
        ----------
        masking_level : str
            The level at which to mask the image. Must be either
            ``"pixel"`` or ``"feature"``.
        std : float
            The standard deviation of the random noise to add to the image.
            Defaults to 1.
        """
        super().__init__(masking_level)
        self.std = std

    def _initialize_baselines(self, samples: torch.Tensor):
        self.baseline = (
            torch.randn(*samples.shape, device=samples.device) * self.std
        )



from tqdm import tqdm
from typing import Dict, Optional, List
from torch import nn
from torch.utils.data import Dataset, DataLoader
from attribench.data import AttributionsDatasetWriter, IndexDataset
from attribench import AttributionMethod
import torch


def compute_attributions(
    model: nn.Module,
    method_dict: Dict[str, AttributionMethod],
    dataset: Dataset,
    batch_size: int,
    writer: Optional[AttributionsDatasetWriter] = None,
    device: Optional[torch.device] = None,
) -> Optional[Dict[str, torch.Tensor]]:
    """Compute attributions for a given model and dataset using a dictionary of 
    attribution methods, and optionally write them to a HDF5 file. If the `writer`
    is `None`, the attributions are simply returned in a dictionary.
    Otherwise, the attributions are written to the HDF5 file and `None` is returned.

    TODO don't write to file, just return the dict

    Parameters
    ----------
    model : nn.Module
        The model for which the attributions should be computed.
    method_dict : Dict[str, AttributionMethod]
        Dictionary of attribution methods.
    dataset : Dataset
        Torch Dataset to use for computing the attributions.
    batch_size : int
        The batch size to use for computing the attributions.
    writer : Optional[AttributionsDatasetWriter], optional
        AttributionsDatasetWriter to write the attributions to, by default `None`.
        If `None`, the attributions are returned in a dictionary.
    device : Optional[torch.device], optional
        Device to use, by default `None`.
        If `None`, the CPU is used.

    Returns
    -------
    Optional[Dict[str, torch.Tensor]]
        If `writer` is `None`, a dictionary of attributions.
    """
    if device is None:
        device = torch.device("cpu")
    model.to(device)
    model.eval()

    index_dataset = IndexDataset(dataset)
    dataloader = DataLoader(
        index_dataset,
        batch_size=batch_size,
        num_workers=4,
        pin_memory=True,
    )

    result_dict: Dict[str, List[torch.Tensor]] = {method_name: [
        torch.zeros(1) for _ in range(len(index_dataset))
    ] for method_name in method_dict.keys()}
    for batch_indices, batch_x, batch_y in tqdm(dataloader):
        batch_x = batch_x.to(device)
        batch_y = batch_y.to(device)
        for method_name, method in method_dict.items():
            with torch.no_grad():
                attrs = method(batch_x, batch_y)
                if writer is None:
                    for idx in batch_indices:
                        result_dict[method_name][idx] = attrs[idx, ...].cpu()
                else:
                    writer.write(
                        batch_indices.cpu().numpy(),
                        attrs.cpu().numpy(),
                        method_name,
                    )
    if writer is None:
        result_dict_cat = {
            method_name: torch.cat(attrs_list)
            for method_name, attrs_list in result_dict.items()
        }
        return result_dict_cat



import pandas as pd
from tqdm import tqdm
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt


class ConvergencePlot:
    """
    Line plot of the median values of the metrics vs. the number of samples.
    Error bars are computed using bootstrapping.
    Allows the user to inspect if metric values have converged,
    i.e. if the benchmark has been run on enough samples.
    """
    def __init__(self, df: pd.DataFrame):
        self.df = df

    def render(self, title=None, bs_samples=1000, interval=5):
        all_medians = []
        for bs_size in tqdm(range(2, self.df.shape[0], interval)):
            medians = []
            for _ in range(bs_samples):
                sample = self.df.sample(n=bs_size, replace=True)
                medians.append(sample.median(axis=0))  # median for each column
            medians = pd.DataFrame(medians)

            medians = pd.melt(medians, var_name="method")
            medians["bs_size"] = bs_size

            all_medians.append(medians)
        all_medians = pd.concat(all_medians)
        fig, ax = plt.subplots(figsize=(20, 20))
        sns.lineplot(
            data=all_medians,
            x="bs_size",
            y="value",
            hue="method",
            estimator=np.median,
            errorbar="ci",
            ax=ax,
        )
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
        ax.set_title(title)
        return fig



import pandas as pd
from krippendorff import krippendorff
from scipy.stats import rankdata
import matplotlib.pyplot as plt
from attribench.plot import Plot
from matplotlib.figure import Figure


class KrippendorffAlphaPlot(Plot):
    """Bar plot of Krippendorff's alpha for each metric."""

    def render(
        self, title: str | None = None, figsize=(10, 10), fontsize=20
    ) -> Figure:
        """Render the plot.

        Parameters
        ----------
        title : str | None, optional
            Title of the figure, by default None
        figsize : tuple, optional
            Size of the figure, by default (10, 10)
        fontsize : int, optional
            Font size of x and y axis ticks, by default 20

        Returns
        -------
        Figure
            Rendered Matplotlib figure.
        """
        k_a = {
            metric_name: krippendorff.alpha(
                rankdata(df.to_numpy(), axis=1), level_of_measurement="ordinal"
            )
            for metric_name, (df, _) in self.dfs.items()
        }
        k_a = pd.DataFrame(k_a, index=["Krippendorff Alpha"]).transpose()
        fig, ax = plt.subplots()
        plt.xticks(fontsize=fontsize)
        plt.yticks(fontsize=fontsize)
        k_a.plot.barh(figsize=figsize, ax=ax)
        if title is not None:
            ax.set_title(title)
        fig.tight_layout()
        return fig



from typing import Tuple, Optional, List
from typing_extensions import override
import pandas as pd
from ._grouped_metric_result import GroupedMetricResult


class InfidelityResult(GroupedMetricResult):
    """Represents results from running the Infidelity metric.
    """
    def __init__(
        self,
        method_names: List[str],
        perturbation_generators: List[str],
        activation_fns: List[str],
        num_samples: int,
    ):
        """
        Parameters
        ----------
        method_names : List[str]
            Names of attribution methods tested by Infidelity.
        perturbation_generators : List[str]
            Names of perturbation generators used by Infidelity.
        activation_fns : List[str]
            Names of activation functions used by Infidelity.
        num_samples : int
            Number of samples on which Infidelity was run.
        """
        levels = {
            "method": method_names,
            "perturbation_generator": perturbation_generators,
            "activation_fn": activation_fns,
        }
        shape = [num_samples, 1]
        level_order = ["method", "perturbation_generator", "activation_fn"]
        super().__init__(method_names, shape, levels, level_order)

    @classmethod
    @override
    def _load(cls, path: str, format="hdf5") -> "InfidelityResult":
        tree = cls._load_tree(path, format)
        res = InfidelityResult(
            tree.levels["method"],
            tree.levels["perturbation_generator"],
            tree.levels["activation_fn"],
            tree.shape[0],
        )
        res.tree = tree
        return res

    def get_df(
        self,
        perturbation_generator: str,
        activation_fn: str,
        methods: Optional[List[str]] = None,
    ) -> Tuple[pd.DataFrame, bool]:
        """Retrieves a dataframe from the result for the given perturbation
        generator and activation function. The dataframe contains a row for
        each method and a column for each sample. Each value is the
        Infidelity for the given method on the given sample.

        Parameters
        ----------
        perturbation_generator : str
            The perturbation generator to use.
        activation_fn : str
            The activation function to use.
        methods : Optional[List[str]], optional
            The methods to include. If None, includes all methods.
            Defaults to None.

        Returns
        -------
        Tuple[pd.DataFrame, bool]
            Dataframe containing results,
            and boolean indicating if higher is better.
        """
        methods = methods if methods is not None else self.method_names
        df_dict = {}
        for method in methods:
            df_dict[method] = self.tree.get(
                method=method,
                perturbation_generator=perturbation_generator,
                activation_fn=activation_fn,
            ).flatten()
        return pd.DataFrame.from_dict(df_dict), False



from typing import Tuple
from .._typing import _check_is_dataset
import h5py
from numpy import typing as npt
import numpy as np


class HDF5DatasetWriter:
    """Class to write a dataset to a HDF5 file.
    Note that this is not implemented in :class:`HDF5Dataset` because
    we want to avoid having the full dataset in memory. This object writes
    the dataset in chunks as they come in.
    """

    def __init__(self, path: str, num_samples: int):
        self.path: str = path
        self.num_samples = num_samples
        self.head = 0

        self.file: h5py.File | None = None
        self.sample_shape: Tuple[int, ...] | None = None

    def init_file(self, sample_shape: Tuple[int, ...]):
        self.sample_shape = sample_shape
        with h5py.File(self.path, "x") as fp:
            fp.create_dataset(
                "samples",
                shape=(self.num_samples, *self.sample_shape),
                dtype=np.float32,
            )
            fp.create_dataset(
                "labels", shape=(self.num_samples,), dtype=np.int64
            )
            self.head = 0

    def write(self, samples: npt.NDArray, labels: npt.NDArray):
        if self.sample_shape is None:
            self.init_file(samples.shape[1:])
        if self.num_samples - self.head < samples.shape[0]:
            raise ValueError("Data size exceeds pre-specified length")
        if samples.shape[0] != labels.shape[0]:
            raise ValueError(
                "Number of samples and number of labels do not match."
            )
        if samples.shape[1:] != self.sample_shape:
            raise ValueError(
                f"Invalid sample shape. Expected: {self.sample_shape}, "
                f"got: {samples.shape[1:]}"
            )
        if self.file is None:
            self.file = h5py.File(self.path, "a")

        samples_dataset = _check_is_dataset(self.file["samples"])
        labels_dataset = _check_is_dataset(self.file["labels"])

        samples_dataset[
            self.head : self.head + samples.shape[0], ...
        ] = samples.astype(np.float32)
        labels_dataset[
            self.head : self.head + labels.shape[0]
        ] = labels.astype(np.int64)
        self.head += samples.shape[0]

    def __del__(self):
        if self.file is not None:
            self.file.close()



import torch
from typing import Callable, List, Mapping, Dict, Union
from attribench.masking import Masker
from attribench.data import AttributionsDataset
from torch import nn
from ._dataset import DeletionDataset
from ._get_predictions import get_predictions
from torch.utils.data import DataLoader
from attribench.result import DeletionResult
from attribench.result._batch_result import BatchResult


def _deletion_batch(
    samples: torch.Tensor,
    labels: torch.Tensor,
    model: Callable,
    attrs: torch.Tensor,
    maskers: Mapping[str, Masker],
    activation_fns: List[str],
    mode: str,
    start: float,
    stop: float,
    num_steps: int,
) -> Dict:
    result_dict = {}
    for masker_name, masker in maskers.items():
        ds = DeletionDataset(
            mode, start, stop, num_steps, samples, attrs, masker
        )
        result_dict[masker_name] = get_predictions(
            ds, labels, model, activation_fns
        )
    return result_dict


def deletion(
    model: nn.Module,
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    maskers: Mapping[str, Masker],
    activation_fns: Union[List[str], str] = "linear",
    mode: str = "morf",
    start: float = 0.0,
    stop: float = 1.0,
    num_steps: int = 100,
    device: torch.device = torch.device("cpu"),
) -> DeletionResult:
    """Computes the Deletion metric for a given :class:`~attribench.data.AttributionsDataset` and model.

    Deletion is computed by iteratively masking the top (Most Relevant First,
    or MoRF) or bottom (Least Relevant First, or LeRF) features of
    the input samples and computing the confidence of the model on the masked
    samples.

    This results in a curve of confidence vs. number of features masked. The
    area under (or equivalently over) this curve is the Deletion metric.

    `start`, `stop`, and `num_steps` are used to determine the range of features
    to mask. The range is determined by `start` and `stop` as a percentage of
    the total number of features. `num_steps` is the number of steps to take
    between `start` and `stop`.

    The Deletion metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`.

    Parameters
    ----------
    model : nn.Module
        Model to compute Deletion on.
    dataset : AttributionsDataset
        Dataset of attributions to compute Deletion on.
    batch_size : int
        Batch size to use when computing model predictions on masked samples.
    maskers : Mapping[str, Masker]
        Dictionary of maskers to use for masking samples.
    activation_fns : Union[List[str], str], optional
        List of activation functions to use when computing model predictions on
        masked samples. If a single string is passed, it is converted to a
        single-element list.
        Default: "linear"
    mode : str, optional
        Mode to use when masking samples. Must be "morf" or "lerf".
        Default: "morf"
    start : float, optional
        Relative start of the range of features to mask. Must be between 0 and 1.
        Default: 0.0
    stop : float, optional
        Relative end of the range of features to mask. Must be between 0 and 1.
        Default: 1.0
    num_steps : int, optional
        Number of steps to use for the range of features to mask.
        Default: 100
    device : torch.device, optional
        Device to use, by default `torch.device("cpu")`

    Returns
    -------
    DeletionResult
    """
    if isinstance(activation_fns, str):
        activation_fns = [activation_fns]

    model.to(device)
    model.eval()

    dataloader = DataLoader(
        attributions_dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    result = DeletionResult(
        attributions_dataset.method_names,
        list(maskers.keys()),
        activation_fns,
        mode,
        attributions_dataset.num_samples,
        num_steps,
    )

    for (
        batch_indices,
        batch_x,
        batch_y,
        batch_attr,
        method_names,
    ) in dataloader:
        batch_x = batch_x.to(device)
        batch_y = batch_y.to(device)
        batch_result = _deletion_batch(
            batch_x,
            batch_y,
            model,
            batch_attr,
            maskers,
            activation_fns,
            mode,
            start,
            stop,
            num_steps,
        )
        result.add(BatchResult(batch_indices, batch_result, method_names))
    return result



from .._metric import Metric
from typing import Dict, Union, Tuple, List
from attribench.data.attributions_dataset._attributions_dataset import (
    AttributionsDataset,
    GroupedAttributionsDataset,
)
from attribench.masking import Masker
from attribench.result._sensitivity_n_result import SensitivityNResult
from ._sensitivity_n_worker import SensitivityNWorker
from ..._worker import WorkerConfig
from attribench._model_factory import ModelFactory


class SensitivityN(Metric):
    """Compute the Sensitivity-n metric for a given :class:`~attribench.data.AttributionsDataset` and model
    using multiple processes.

    Sensitivity-n is computed by iteratively masking a random subset of `n` features
    of the input samples and computing the output of the model on the masked
    samples.

    For each random subset of masked features, the sum of the attributions is
    also computed. This results in two series of values: the model output and
    the sum of the attributions. The Sensitivity-n metric is the correlation
    between these two series.

    This is repeated for different values of `n` between `min_subset_size` and
    `max_subset_size` in `num_steps` steps. `min_subset_size` and `max_subset_size`
    are percentages of the total number of features.
    For each value of `n`, `num_subsets` random subsets are generated.

    If segmented is True, then the Seg-Sensitivity-n metric is computed.
    This metric is analogous to Sensitivity-n, but instead of using random
    subsets of features, the images are first segmented into superpixels and
    then random subsets of superpixels are masked. This improves the
    signal-to-noise ratio of the metric for high-resolution images.

    The Sensitivity-n metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`. The number of processes is
    determined by the number of devices. If `devices` is None, then all
    available devices are used. Samples are distributed evenly across the
    processes.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        maskers: Dict[str, Masker],
        activation_fns: Union[List[str], str],
        min_subset_size: float,
        max_subset_size: float,
        num_steps: int,
        num_subsets: int,
        segmented=False,
        address="localhost",
        port="12355",
        devices: Tuple | None = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        attributions_dataset : AttributionsDataset
            Dataset containing the attributions to compute Sensitivity-n on.
        batch_size : int
            Batch size to use when computing the model output.
        maskers : Dict[str, Masker]
            Dictionary of maskers to use. Keys are the names of the maskers.
        activation_fns : Union[Tuple[str], str]
            Activation functions to use. If a single string is passed, then the
            it is converted to a single-element list.
        min_subset_size : float
            Minimum percentage of features to mask.
        max_subset_size : float
            Maximum percentage of features to mask.
        num_steps : int
            Number of steps between `min_subset_size` and `max_subset_size`.
        num_subsets : int
            Number of random subsets to generate for each value of `n`.
        segmented : bool
            If True, then the Seg-Sensitivity-n metric is computed.
        address : str, optional
            Address to use for the distributed computation.
            Defaults to "localhost".
        port : str | int, optional
            Port to use for the distributed computation.
            Defaults to "12355".
        devices : Tuple | None
            Devices to use for the distributed computation. If None, then all
            available devices are used.
        """
        super().__init__(
            model_factory,
            attributions_dataset,
            batch_size,
            address,
            port,
            devices,
        )
        if isinstance(activation_fns, str):
            activation_fns = [activation_fns]
        self.activation_fns: List[str] = activation_fns
        self.dataset = GroupedAttributionsDataset(attributions_dataset)
        self.maskers = maskers
        self.num_subsets = num_subsets
        self.num_steps = num_steps
        self.max_subset_size = max_subset_size
        self.min_subset_size = min_subset_size
        self.segmented = segmented
        self._result = SensitivityNResult(
            attributions_dataset.method_names,
            list(maskers.keys()),
            self.activation_fns,
            attributions_dataset.num_samples,
            num_steps,
        )

    def _create_worker(
        self, worker_config: WorkerConfig
    ) -> SensitivityNWorker:
        return SensitivityNWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.min_subset_size,
            self.max_subset_size,
            self.num_steps,
            self.num_subsets,
            self.maskers,
            self.activation_fns,
            self.segmented,
        )



from ._metric_result import MetricResult
from ._grouped_metric_result import GroupedMetricResult
from ._deletion_result import DeletionResult
from ._insertion_result import InsertionResult
from ._impact_coverage_result import ImpactCoverageResult
from ._infidelity_result import InfidelityResult
from ._max_sensitivity_result import MaxSensitivityResult
from ._minimal_subset_result import MinimalSubsetResult
from ._sensitivity_n_result import SensitivityNResult
from ._parameter_randomization_result import ParameterRandomizationResult


from attribench.masking.image import ImageMasker
import torch


class ConstantImageMasker(ImageMasker):
    """Image masker that masks pixels or features by replacing them with
    a given constant value.
    """
    def __init__(self, masking_level: str, mask_value=0.0):
        """
        Parameters
        ----------
        feature_level : str
            The level at which to mask the image. Must be either
            ``"pixel"`` or ``"feature"``.
        mask_value : float
            The value to use for masking. Defaults to 0.0.
        """
        super().__init__(masking_level)
        self.mask_value = mask_value

    def _initialize_baselines(self, samples: torch.Tensor):
        self.baseline = (
            torch.ones(samples.shape, device=samples.device) * self.mask_value
        )



from torch import nn
import torch
from typing import Dict, List
from attribench.data.attributions_dataset._attributions_dataset import (
    GroupedAttributionsDataset,
    AttributionsDataset,
)
from ._perturbation_generator import PerturbationGenerator
from torch.utils.data import DataLoader
from attribench._activation_fns import ACTIVATION_FNS
from attribench.result._infidelity_result import InfidelityResult
from attribench.result._grouped_batch_result import GroupedBatchResult


def _infidelity_batch(
    model: nn.Module,
    batch_x: torch.Tensor,
    batch_y: torch.Tensor,
    batch_attr: Dict[str, torch.Tensor],
    perturbation_generators: Dict[str, PerturbationGenerator],
    num_perturbations: int,
    activation_fns: List[str],
    device: torch.device,
):
    batch_x = batch_x.to(device)
    batch_y = batch_y.to(device)
    method_names: List[str] = list(batch_attr.keys())

    # method_name -> perturbation_generator
    # -> activation_fn -> [batch_size, 1]
    batch_result: Dict[str, Dict[str, Dict[str, torch.Tensor]]] = {
        method_name: {
            pg_name: {
                activation_fn: torch.zeros(1)
                for activation_fn in activation_fns
            }
            for pg_name in perturbation_generators.keys()
        }
        for method_name in method_names
    }

    # Attributions need to be moved to the device,
    # because we will be computing dot products between
    # attributions and perturbations later.
    # They also need to have the same shape as the samples for this.
    # Any axis that has length 1 in the attributions is repeated to
    # match the sample shape
    tensor_attributions: Dict[str, torch.Tensor] = {}
    for attribution_method, attributions in batch_attr.items():
        for axis in range(len(attributions.shape)):
            if attributions.shape[axis] == 1:
                attributions = torch.repeat_interleave(
                    attributions, batch_x.shape[axis], dim=axis
                )
        tensor_attributions[attribution_method] = attributions.flatten(1).to(
            device
        )

    # Get original model output on the samples
    # (dict: activation_fn -> torch.Tensor)
    orig_output = {}
    with torch.no_grad():
        for fn in activation_fns:
            # [batch_size, 1]
            orig_output[fn] = ACTIVATION_FNS[fn](model(batch_x)).gather(
                dim=1, index=batch_y.unsqueeze(-1)
            )

    for (
        pert_name,
        pert_generator,
    ) in perturbation_generators.items():
        pert_generator.set_samples(batch_x)
        dot_products: Dict[str, List[torch.Tensor]] = {
            method: [] for method in batch_attr.keys()
        }
        pred_diffs = {afn: [] for afn in activation_fns}

        for _ in range(num_perturbations):
            # Get perturbation vector I and perturbed samples (x - I)
            perturbation_vector = pert_generator.generate_perturbation()
            perturbed_x = batch_x - perturbation_vector

            # Get output of model on perturbed sample
            with torch.no_grad():
                perturbed_output = model(perturbed_x)

            # Save the prediction difference and perturbation vector
            for fn in activation_fns:
                activated_perturbed_output = ACTIVATION_FNS[fn](
                    perturbed_output
                ).gather(dim=1, index=batch_y.unsqueeze(-1))
                pred_diffs[fn].append(
                    torch.squeeze(orig_output[fn] - activated_perturbed_output)
                )

            # Compute dot products of perturbation vectors with all
            # attributions for each sample
            for (
                attribution_method,
                attributions,
            ) in tensor_attributions.items():
                # (batch_size)
                dot_products[attribution_method].append(
                    (perturbation_vector.flatten(1) * attributions).sum(dim=-1)
                )

        # For each method and activation function, compute infidelity
        # activation_fn -> [num_perturbations, batch_size]
        tensor_pred_diffs = {
            afn: torch.stack(pred_diffs[afn], dim=0)
            for afn in pred_diffs.keys()
        }
        for method in tensor_attributions.keys():
            # Dot prodcts for this method
            method_dot_products = torch.stack(
                dot_products[method]
            )  # [num_perturbations, batch_size]
            # Denominator for normalizing constant beta
            beta_denominator = torch.mean(
                method_dot_products**2, dim=0, keepdim=True
            )  # [1, batch_size]
            for afn in activation_fns:
                # Numerator for normalizing constant beta depends on
                # activation function
                # [1, batch_size]
                beta_numerator = torch.mean(
                    method_dot_products * tensor_pred_diffs[afn],
                    dim=0,
                    keepdim=True,
                )
                beta = beta_numerator / beta_denominator
                # If attribution map is constant 0,
                # dot products will be 0 and beta will be nan. Set to 0.
                beta[torch.isnan(beta)] = 0
                # [batch_size, 1]
                infidelity = torch.mean(
                    (beta * method_dot_products - tensor_pred_diffs[afn]) ** 2,
                    dim=0,
                ).unsqueeze(-1)
                batch_result[method][pert_name][afn] = (
                    infidelity.cpu().detach().numpy()
                )
    return batch_result


def infidelity(
    model: nn.Module,
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    activation_fns: List[str],
    perturbation_generators: Dict[str, PerturbationGenerator],
    num_perturbations: int,
    device: torch.device = torch.device("cpu"),
):
    """Computes the Infidelity metric for a given :class:`~attribench.data.AttributionsDataset` and model.

    Infidelity is computed by generating perturbations for each sample in the
    dataset and computing the difference in the model's output on the original
    sample and the perturbed sample. This difference is then compared to the
    dot product of the perturbation vector and the attribution map for each
    attribution method. The Infidelity metric is the mean squared error between
    these two values.

    The idea is that if the dot product is large, then the perturbation vector
    is aligned with the attribution map, and the model's output should change
    significantly when the perturbation is applied. If the dot product is small,
    then the perturbation vector is not aligned with the attribution map, and
    the model's output should not change significantly when the perturbation is
    applied.

    The mean squared error is computed for `num_perturbations` perturbations
    for each sample. The `perturbation_generators` argument is a dictionary
    mapping perturbation generator names to `PerturbationGenerator` objects.
    These objects can be used to implement different versions of Infidelity.

    The Infidelity metric is computed for each perturbation generator in
    `perturbation_generators` and each activation function in `activation_fns`.

    Parameters
    ----------
    model : nn.Module
        Model to compute Infidelity on.
    attributions_dataset : AttributionsDataset
        Dataset of attributions to compute Infidelity on.
    batch_size : int
        Batch size to use when computing Infidelity.
    perturbation_generators : Dict[str, PerturbationGenerator]
        Dictionary of perturbation generators to use for generating
        perturbations.
    num_perturbations : int
        Number of perturbations to generate for each sample.
    activation_fns : Tuple[str]
        Tuple of activation functions to use when computing Infidelity.
    device : torch.device, optional
        Device to use, by default `torch.device("cpu")`
    """
    grouped_dataset = GroupedAttributionsDataset(attributions_dataset)
    dataloader = DataLoader(
        grouped_dataset, batch_size=batch_size, num_workers=4
    )
    result = InfidelityResult(
        attributions_dataset.method_names,
        list(perturbation_generators.keys()),
        activation_fns,
        num_samples=attributions_dataset.num_samples,
    )
    for batch_indices, batch_x, batch_y, batch_attr in dataloader:
        batch_result = _infidelity_batch(
            model,
            batch_x,
            batch_y,
            batch_attr,
            perturbation_generators,
            num_perturbations,
            activation_fns,
            device,
        )
        result.add(GroupedBatchResult(batch_indices, batch_result))



from attribench.masking import ImageMasker
import torch


class SampleAverageMasker(ImageMasker):
    def __init__(self, feature_level):
        super().__init__(feature_level)

    def _initialize_baselines(self, samples: torch.tensor):
        batch_size, num_channels, rows, cols = samples.shape
        self.baseline = (
            torch.mean(samples.flatten(2), dim=-1)
            .reshape(batch_size, num_channels, 1, 1)
            .repeat(1, 1, rows, cols)
        )



from typing_extensions import override
import os
import yaml
import h5py
import numpy as np
from numpy import typing as npt
from typing import List, Tuple, Optional
from attribench.data.nd_array_tree._random_access_nd_array_tree import (
    RandomAccessNDArrayTree,
)
from ._metric_result import MetricResult
import pandas as pd


def _aoc(x: np.ndarray, columns: Optional[npt.NDArray] = None):
    if columns is not None:
        x = x[..., columns]
    return x[..., 0] - _auc(x, columns)


def _auc(x: np.ndarray, columns: Optional[npt.NDArray] = None):
    # TODO do we have to normalize by the first value? Same for AOC.
    # Should actually not make a difference.
    if columns is not None:
        x = x[..., columns]
    l = x.shape[-1] if columns is None else columns.shape[0]
    return np.sum(x, axis=-1) / l


class DeletionResult(MetricResult):
    """
    Represents results from running the Deletion metric.
    """

    def __init__(
        self,
        method_names: List[str],
        maskers: List[str],
        activation_fns: List[str],
        mode: str,
        num_samples: int,
        num_steps: int,
    ):
        """
        Parameters
        ----------
        method_names : List[str]
            Names of attribution methods tested by Deletion.
        maskers : List[str]
            Names of maskers used by Deletion.
        activation_fns : List[str]
            Names of activation functions used by Deletion.
        mode : str
            Indicates if Deletion-MoRF or Deletion-LeRF was used.
            Options: "morf", "lerf"
        num_samples : int
            Number of samples on which Deletion was run.
        num_steps : int
            Number of steps used by Deletion.
        """
        levels = {
            "method": method_names,
            "masker": maskers,
            "activation_fn": activation_fns,
        }
        level_order = ["method", "masker", "activation_fn"]
        shape = [num_samples, num_steps]
        super().__init__(method_names, shape, levels, level_order)
        self.mode = mode

    @override
    def save(self, path: str, format="hdf5"):
        super().save(path, format)

        # Save additional metadata
        if format == "hdf5":
            with h5py.File(path, mode="a") as fp:
                fp.attrs["mode"] = self.mode
        elif format == "csv":
            with open(os.path.join(path, "metadata.yaml"), "r") as fp:
                metadata = yaml.safe_load(fp)
            metadata["mode"] = self.mode
            with open(os.path.join(path, "metadata.yaml"), "w") as fp:
                yaml.dump(metadata, fp)

    @classmethod
    def _load_tree_mode(
        cls, path: str, format="hdf5"
    ) -> Tuple[RandomAccessNDArrayTree, str]:
        """Loads the tree and mode from a file or directory.

        Parameters
        ----------
        path : str
            Path to the file or directory.
        format : str, optional
            Format of the saved result.
            Options: "hdf5", "csv".
            By default "hdf5".

        Returns
        -------
        Tuple[RandomAccessNDArrayTree, str]
            The RandomAccessNDArrayTree object and the mode as a string.

        Raises
        ------
        ValueError
            If the format argument is not valid.
        """
        if format == "hdf5":
            with h5py.File(path, "r") as fp:
                tree = RandomAccessNDArrayTree.load_from_hdf(fp)
                mode = str(fp.attrs["mode"])
        elif format == "csv":
            with open(os.path.join(path, "metadata.yaml"), "r") as fp:
                metadata = yaml.safe_load(fp)
            tree = RandomAccessNDArrayTree.load_from_dir(path)
            mode = metadata["mode"]
        else:
            raise ValueError("Invalid format", format)
        return tree, mode

    @classmethod
    @override
    def _load(cls, path: str, format="hdf5") -> "DeletionResult":
        tree, mode = cls._load_tree_mode(path, format)
        res = DeletionResult(
            tree.levels["method"],
            tree.levels["masker"],
            tree.levels["activation_fn"],
            mode,
            tree.shape[0],
            tree.shape[1],
        )
        res.tree = tree
        return res

    @override
    def get_df(
        self,
        masker: str,
        activation_fn: str,
        agg_fn="auc",
        methods: Optional[List[str]] = None,
        columns: Optional[npt.NDArray] = None,
    ) -> Tuple[pd.DataFrame, bool]:
        """
        Retrieves a dataframe from the result for a given masker and
        activation function. The dataframe contains a row for each sample and a
        column for each method. Each value is the AUC/AOC for the given method
        on the given sample.

        Parameters
        ----------
        masker : str
            The masker to use.
        activation_fn : str
            The activation function to use.
        agg_fn : str
            Either "auc" for AUC or "aoc" for AOC.
        methods : Optional[List[str]]
            The methods to include. If None, includes all methods.
        columns : Optional[npt.NDArray]
            The columns used in the AUC/AOC calculation.
            If None, uses all columns.

        Returns
        -------
        Tuple[pd.DataFrame, bool]
            dataframe containing results,
            and boolean indicating if higher is better.
        """
        higher_is_better = (self.mode == "morf" and agg_fn == "aoc") or (
            self.mode == "lerf" and agg_fn == "auc"
        )
        methods = methods if methods is not None else self.method_names
        df_dict = {}
        agg_fns = {"auc": _auc, "aoc": _aoc}
        for method in methods:
            array = self.tree.get(
                masker=masker, activation_fn=activation_fn, method=method
            )
            df_dict[method] = agg_fns[agg_fn](array, columns)
        return pd.DataFrame.from_dict(df_dict), higher_is_better



import numpy as np
from torch import nn
import torch
import numpy.typing as npt
from typing import Callable, List, Mapping, Dict, Tuple
from attribench.masking import Masker
from attribench.masking.image import ImageMasker
from torch.utils.data import DataLoader
from attribench.data import AttributionsDataset
from attribench._activation_fns import ACTIVATION_FNS
from ._dataset import SensitivityNDataset, SegSensNDataset
from attribench._segmentation import segment_attributions
from attribench._stat import rowwise_pearsonr
from attribench.result import SensitivityNResult
from attribench.result._grouped_batch_result import GroupedBatchResult
from attribench.data.attributions_dataset._attributions_dataset import (
    GroupedAttributionsDataset,
)


def _get_orig_output(
    samples: torch.Tensor, model: Callable, activation_fns: List[str]
):
    activated_orig_output = {}
    with torch.no_grad():
        orig_output = model(samples)
        for activation_fn in activation_fns:
            activated_orig_output[activation_fn] = ACTIVATION_FNS[
                activation_fn
            ](orig_output)
    return activated_orig_output


def _compute_out_diffs(
    model: Callable,
    ds: SensitivityNDataset | SegSensNDataset,
    activation_fns: List[str],
    orig_output: Dict[str, torch.Tensor],
    labels: torch.Tensor,
) -> Tuple[Dict[str, Dict[int, npt.NDArray]], Dict[int, npt.NDArray]]:
    n_range = ds.n_range
    output_diff_shape = (ds.samples.shape[0], ds.num_subsets)
    # Calculate differences in output and removed indices
    # (will be re-used for all methods)
    # activation_fn -> n -> [batch_size, num_subsets]
    output_diffs: Dict[str, Dict[int, npt.NDArray]] = {
        activation_fn: {n: np.zeros(output_diff_shape) for n in n_range}
        for activation_fn in activation_fns
    }
    removed_indices: Dict[int, npt.NDArray] = {
        n: np.zeros((ds.samples.shape[0], ds.num_subsets, n), dtype=int)
        for n in n_range
    }
    # TODO why do we not use a dataloader here?
    for i in range(len(ds)):
        batch, indices, n, subset_idx = ds[i]
        n = n.item()
        with torch.no_grad():
            output = model(batch)
        for activation_fn in activation_fns:
            activated_output = ACTIVATION_FNS[activation_fn](output)
            # [batch_size, 1]
            output_diffs[activation_fn][n][:, subset_idx] = (
                (orig_output[activation_fn] - activated_output)
                .gather(dim=1, index=labels.unsqueeze(-1))
                .flatten()
                .detach()
                .cpu()
                .numpy()
            )
        removed_indices[n][:, subset_idx, :] = indices  # [batch_size, n]
    return output_diffs, removed_indices


def _compute_correlations(
    method_names: List[str],
    batch_attr: Dict[str, torch.Tensor],
    ds: SensitivityNDataset | SegSensNDataset,
    segmented: bool,
    removed_indices: Dict[int, npt.NDArray],
    output_diffs: Dict[str, Dict[int, npt.NDArray]],
    activation_fns: List[str],
) -> Dict[str, Dict[str, torch.Tensor]]:
    # activation_fn -> method_name -> [batch_size, len(n_range)]
    result = {
        activation_fn: {
            method_name: torch.zeros((ds.samples.shape[0], len(ds.n_range)))
            for method_name in method_names
        }
        for activation_fn in activation_fns
    }
    # Compute correlations for all methods
    # TODO can we use joblib here to compute this in parallel?
    for method_name in method_names:
        attrs = batch_attr[method_name].cpu().numpy()
        if segmented:
            assert isinstance(ds, SegSensNDataset)
            attrs = segment_attributions(
                ds.segmented_images.cpu().numpy(), attrs
            )
        # [batch_size, 1, -1]
        attrs = attrs.reshape((attrs.shape[0], 1, -1))
        for n_idx, n in enumerate(ds.n_range):
            # [batch_size, num_subsets, n]
            n_mask_attrs = np.take_along_axis(
                attrs, axis=-1, indices=removed_indices[n]
            )
            for activation_fn in activation_fns:
                # Compute sum of attributions
                # [batch_size, num_subsets]
                n_sum_of_attrs = n_mask_attrs.sum(axis=-1)
                n_output_diffs = output_diffs[activation_fn][n]
                # Compute correlation between output difference and
                # sum of attribution values
                result[activation_fn][method_name][:, n_idx] = torch.tensor(
                    rowwise_pearsonr(n_sum_of_attrs, n_output_diffs)
                )
    return result


def _sens_n_batch(
    samples: torch.Tensor,
    labels: torch.Tensor,
    model: Callable,
    attrs: Dict[str, torch.Tensor],
    maskers: Mapping[str, Masker],
    activation_fns: List[str],
    n_range: npt.NDArray,
    num_subsets: int,
    segmented: bool,
) -> Dict[str, Dict[str, Dict[str, torch.Tensor]]]:
    method_names = list(attrs.keys())
    orig_output = _get_orig_output(samples, model, activation_fns)
    # masker_name -> activation_fn -> method_name -> [batch_size, num_steps]
    batch_result: Dict[str, Dict[str, Dict[str, torch.Tensor]]] = {}

    for masker_name, masker in maskers.items():
        # Create pseudo-dataset to generate perturbed samples
        if segmented:
            ds = SegSensNDataset(n_range, num_subsets, samples)
            assert isinstance(masker, ImageMasker)
            ds.set_masker(masker)
        else:
            ds = SensitivityNDataset(n_range, num_subsets, samples, masker)

        output_diffs, removed_indices = _compute_out_diffs(
            model, ds, activation_fns, orig_output, labels
        )

        batch_result[masker_name] = _compute_correlations(
            method_names,
            attrs,
            ds,
            segmented,
            removed_indices,
            output_diffs,
            activation_fns,
        )
    return batch_result


def sensitivity_n(
    model: nn.Module,
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    maskers: Mapping[str, Masker],
    activation_fns: str | List[str],
    min_subset_size: float,
    max_subset_size: float,
    num_steps: int,
    num_subsets: int,
    segmented: bool,
    device: torch.device = torch.device("cpu"),
) -> SensitivityNResult:
    """Computes the Sensitivity-n metric for a given :class:`~attribench.data.AttributionsDataset` and model.

    Sensitivity-n is computed by iteratively masking a random subset of `n` features
    of the input samples and computing the output of the model on the masked
    samples.

    For each random subset of masked features, the sum of the attributions is
    also computed. This results in two series of values: the model output and
    the sum of the attributions. The Sensitivity-n metric is the correlation
    between these two series.

    This is repeated for different values of `n` between `min_subset_size` and
    `max_subset_size` in `num_steps` steps. `min_subset_size` and `max_subset_size`
    are percentages of the total number of features.
    For each value of `n`, `num_subsets` random subsets are generated.

    If segmented is True, then the Seg-Sensitivity-n metric is computed.
    This metric is analogous to Sensitivity-n, but instead of using random
    subsets of features, the images are first segmented into superpixels and
    then random subsets of superpixels are masked. This improves the
    signal-to-noise ratio of the metric for high-resolution images.

    The Sensitivity-n metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`.

    Parameters
    ----------
    model : nn.Module
        Model to compute Sensitivity-n for.
    attributions_dataset : AttributionsDataset
        Dataset containing the attributions to compute Sensitivity-n on.
    batch_size : int
        Batch size to use when computing model output on masked samples.
    maskers : Dict[str, Masker]
        Dictionary of maskers to use. Keys are the names of the maskers.
    activation_fns : Union[Tuple[str], str]
        Activation functions to use. If a single string is passed, then the
        it is converted to a single-element list.
    min_subset_size : float
        Minimum percentage of features to mask.
    max_subset_size : float
        Maximum percentage of features to mask.
    num_steps : int
        Number of steps between `min_subset_size` and `max_subset_size`.
    num_subsets : int
        Number of random subsets to generate for each value of `n`.
    segmented : bool
        If True, then the Seg-Sensitivity-n metric is computed.
    device : torch.device, optional
        Device to use, by default torch.device("cpu")

    Returns
    -------
    SensitivityNResult
    """
    if isinstance(activation_fns, str):
        activation_fns = [activation_fns]

    model.to(device)
    model.eval()

    grouped_dataset = GroupedAttributionsDataset(attributions_dataset)
    dataloader = DataLoader(
        grouped_dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    result = SensitivityNResult(
        attributions_dataset.method_names,
        list(maskers.keys()),
        list(activation_fns),
        num_samples=attributions_dataset.num_samples,
        num_steps=num_steps,
    )

    # Compute range of subset sizes
    n_range = np.linspace(min_subset_size, max_subset_size, num_steps)
    if segmented:
        n_range = n_range * 100
    else:
        total_num_features = np.prod(attributions_dataset.attributions_shape)
        n_range = n_range * total_num_features
    n_range = n_range.astype(int)

    for (
        batch_indices,
        batch_x,
        batch_y,
        batch_attr,
    ) in dataloader:
        batch_x = batch_x.to(device)
        batch_y = batch_y.to(device)
        batch_result = _sens_n_batch(
            batch_x,
            batch_y,
            model,
            batch_attr,
            maskers,
            activation_fns,
            n_range,
            num_subsets,
            segmented,
        )
        result.add(GroupedBatchResult(batch_indices, batch_result))
    return result



from typing import Tuple, Optional, List
import os
import yaml
from typing_extensions import override

import h5py
import pandas as pd

from ._metric_result import MetricResult
from attribench.data.nd_array_tree._random_access_nd_array_tree import (
    RandomAccessNDArrayTree,
)


class MinimalSubsetResult(MetricResult):
    """Represents results from running the MinimalSubset metric."""

    def __init__(
        self,
        method_names: List[str],
        maskers: List[str],
        mode: str,
        num_samples: int,
    ):
        """
        Parameters
        ----------
        method_names : List[str]
            Names of attribution methods tested by MinimalSubset.
        maskers : List[str]
            Names of maskers used by MinimalSubset.
        mode : str
            Indicates if Minimal Subset Deletion or Minimal Subset Insertion
            was used.
            Options: "deletion", "insertion"
        num_samples : int
            Number of samples on which MinimalSubset was run.
        """
        levels = {"method": method_names, "masker": maskers}
        level_order = ["method", "masker"]
        shape = [num_samples, 1]
        super().__init__(method_names, shape, levels, level_order)
        self.mode = mode

    def save(self, path: str, format="hdf5"):
        super().save(path, format)

        if format == "hdf5":
            with h5py.File(path, mode="a") as fp:
                fp.attrs["mode"] = self.mode
        elif format == "csv":
            with open(os.path.join(path, "metadata.yaml"), "r") as fp:
                metadata = yaml.safe_load(fp)
            metadata["mode"] = self.mode
            with open(os.path.join(path, "metadata.yaml"), "w") as fp:
                yaml.dump(metadata, fp)

    @classmethod
    def _load_tree_mode(
        cls, path: str, format="hdf5"
    ) -> Tuple[RandomAccessNDArrayTree, str]:
        if format == "hdf5":
            with h5py.File(path, "r") as fp:
                tree = RandomAccessNDArrayTree.load_from_hdf(fp)
                mode = str(fp.attrs["mode"])
        elif format == "csv":
            with open(os.path.join(path, "metadata.yaml"), "r") as fp:
                metadata = yaml.safe_load(fp)
            tree = RandomAccessNDArrayTree.load_from_dir(path)
            mode = metadata["mode"]
        else:
            raise ValueError("Invalid format", format)
        return tree, mode

    @classmethod
    @override
    def _load(cls, path: str, format="hdf5") -> "MinimalSubsetResult":
        tree, mode = cls._load_tree_mode(path, format)
        res = MinimalSubsetResult(
            tree.levels["method"],
            tree.levels["masker"],
            mode,
            tree.shape[0],
        )
        res.tree = tree
        return res

    @override
    def get_df(
        self, masker: str, methods: Optional[List[str]] = None
    ) -> Tuple[pd.DataFrame, bool]:
        """Retrieves a dataframe from the result for the given masker. The
        dataframe contains a row for each method and a column for each sample.
        Each value is the MinimalSubset for the given method on the given
        sample.

        Parameters
        ----------
        masker : str
            The masker to include.
        methods : Optional[List[str]], optional
            The methods to include. If None, includes all methods.
            Defaults to None.
        """
        higher_is_better = False
        methods = methods if methods is not None else self.method_names
        df_dict = {
            method: self.tree.get(masker=masker, method=method).flatten()
            for method in methods
        }
        return pd.DataFrame.from_dict(df_dict), higher_is_better



import torch
from torch.utils.data import TensorDataset
from attribench.data._index_dataset import IndexDataset
from .._typing import _check_is_dataset
from torch.utils.data import Dataset
import numpy as np
import h5py
from typing import List, Dict, Tuple


def _max_abs(arr: torch.Tensor, dim: int) -> torch.Tensor:
    return torch.max(torch.abs(arr), dim=dim, keepdim=True)


def _mean(arr: torch.Tensor, dim: int) -> torch.Tensor:
    return torch.mean(arr, dim=dim, keepdim=True)


def _check_is_dataset_or_tensor(obj) -> h5py.Dataset | torch.Tensor:
    if isinstance(obj, h5py.Dataset) or isinstance(obj, torch.Tensor):
        return obj
    else:
        raise ValueError(
            f"Expected obj to be a Dataset or Tensor, but got {type(obj)}"
        )


def _get_attributions_shape(
    attributions: Dict[str, torch.Tensor] | h5py.File, method_names: List[str]
) -> Tuple[int, ...]:
    shape = None
    for m_name in method_names:
        if shape is None:
            # If shape is None, set it to the shape of the first method
            if isinstance(attributions, h5py.File):
                dataset = _check_is_dataset(attributions[m_name])
                shape = dataset.shape
            else:
                shape = attributions[m_name].shape
        else:
            # Otherwise, check if the shape for the current method
            # is the same as the first
            if isinstance(attributions, h5py.File):
                dataset = _check_is_dataset(attributions[m_name])
                cur_shape = dataset.shape
            else:
                cur_shape = attributions[m_name].shape
            if shape != cur_shape:
                raise ValueError(
                    "Attributions must have the same shape for each method"
                )
    if shape is None:
        raise ValueError("Attributions must not be empty")
    return shape


def _parse_attributions_dict(
    attributions: Dict[str, torch.Tensor],
    methods: List[str] | None,
) -> Tuple[List[str], int, Tuple[int, ...]]:
    if methods is None:
        method_names = list(attributions.keys())
    elif all(m in attributions.keys() for m in methods):
        method_names = methods
    else:
        raise ValueError(f"Invalid methods: {methods}")

    shape = _get_attributions_shape(attributions, method_names)

    num_samples = shape[0]
    attributions_shape = shape[1:]
    return method_names, num_samples, attributions_shape


def _parse_attributions_file(
    path: str, methods: List[str] | None
) -> Tuple[List[str], int, Tuple[int, ...]]:
    with h5py.File(path, "r") as fp:
        # Check if methods argument is valid
        if methods is None:
            method_names = list(fp.keys())
        elif all(m in fp for m in methods):
            method_names = methods
        else:
            raise ValueError(f"Invalid methods: {methods}")

        # Check if num_samples metadata is valid
        num_samples = fp.attrs["num_samples"]
        if isinstance(num_samples, np.integer):
            num_samples = int(num_samples)
        else:
            raise ValueError(
                f"Expected num_samples to be an integer,"
                f" but got {type(num_samples)}"
            )

        # Check if attributions have the same shape for each method
        shape = _get_attributions_shape(fp, method_names)
        attributions_shape = shape[1:]
    return method_names, num_samples, attributions_shape


class AttributionsDataset(IndexDataset):
    """
    Represents a dataset containing attributions for a set of samples and
    attribution methods.

    The samples and labels can be given in two ways. Either a PyTorch
    ``Dataset`` is passed to the ``samples`` argument
    containing both the samples and the labels, or a Tensor is passed to the
    ``samples`` argument and a Tensor is passed to the ``labels`` argument.

    An AttributionsDataset can be constructed from a dictionary of attributions
    or from an HDF5 file containing the attributions. 
    If attributions are given using a dictionary, the dictionary must map
    method names to Tensors containing the attributions for each sample. The
    attributions must have the same shape for each method. The shape of the
    attributions must be ``[num_samples, *sample_shape]``.

    If attributions are given using an HDF5 file, the file must contain a
    dataset for each attribution method. The dataset must have the same shape
    for each method. The shape of the dataset must be
    ``[num_samples, *sample_shape]``. The file must also contain an attribute
    ``num_samples`` specifying the number of samples in the dataset.

    A list of method names can be given using the ``methods`` argument. If
    ``methods`` is None, all methods in the attributions dictionary or file
    are used. Otherwise, only the methods in the ``methods`` list are used.

    Attributions can be aggregated over some dimension by specifying the
    aggregate_dim and aggregate_method arguments. The aggregate_dim argument
    specifies the dimension over which to aggregate. The aggregate_method
    argument specifies the method to use for aggregation. The aggregate_method
    argument must be one of ``"mean"`` or ``"max_abs"``. Note that the
    aggregate_dim argument is specified in terms of the shape of the
    attributions, i.e. excluding the ``num_samples`` dimension.
    
    For example, if the
    attributions have shape ``[num_samples, 3, 32, 32]``, then the
    attributions can be aggregated over the channel dimension by setting
    ``aggregate_dim=0``. The resulting attributions will have shape
    ``[num_samples, 32, 32]``.
    """

    def __init__(
        self,
        samples: Dataset | torch.Tensor,
        labels: torch.Tensor | None = None,
        path: str | None = None,
        attributions: Dict[str, torch.Tensor] | None = None,
        methods: List[str] | None = None,
        aggregate_dim: int = 0,
        aggregate_method: str | None = None,
    ):
        """
        Parameters
        ----------
        samples: Dataset | torch.Tensor
            A Dataset containing samples and labels, or a Tensor containing the
            samples for which attributions are given.
        labels: torch.Tensor | None
            A Tensor containing the labels for the samples.
            Only used if samples is a Tensor.
        path: str | None
            Path to an HDF5 file containing the attributions.
            If None, attributions must be given as a dictionary.
        attributions: Dict[str, torch.Tensor] | None
            A dictionary mapping attribution method names to Tensors containing
            the attributions for each sample. If None, a path to an HDF5 file
            must be given.
        methods: List[str] | None
            A list of method names to use. If None, all methods in the
            attributions dictionary are used.
        aggregate_dim: int
            If not None, aggregate the attributions over the given dimension.
        aggregate_method: str | None
            If not None, aggregate the attributions using the given method.
            Must be one of "mean" or "max_abs" or None.
        
        Raises
        ------
        ValueError
            If attributions is None and path is None, or if labels is None and
            samples is a Tensor.
        """
        self.path = path

        # If samples and labels are given as Tensors, wrap them in a Dataset
        self.samples_dataset: Dataset
        if isinstance(samples, torch.Tensor):
            if labels is None:
                raise ValueError(
                    "Labels must be given if samples are given as a Tensor"
                )
            self.samples_dataset = TensorDataset(samples, labels)
        else:
            self.samples_dataset = samples

        # Handle attributions dict or file
        self.attributions: Dict[str, torch.Tensor] | h5py.File | None = None
        self.method_names: List[str]
        orig_attributions_shape: Tuple[int, ...]
        if attributions is not None:
            # If attributions are given as a dict, parse the dict for metadata.
            (
                self.method_names,
                self.num_samples,
                orig_attributions_shape,
            ) = _parse_attributions_dict(attributions, methods)
            self.attributions = attributions
        else:
            # Otherwise, a path must be given. Load metadata from HDF5 file.
            if path is None:
                raise ValueError("Either attributions or path must be given")
            (
                self.method_names,
                self.num_samples,
                orig_attributions_shape,
            ) = _parse_attributions_file(path, methods)

        # Handle aggregation if necessary
        self.aggregate_fn = None
        self.aggregate_dim = aggregate_dim
        self.attributions_shape: Tuple[int, ...]
        if aggregate_method is not None:
            agg_fns = {"mean": _mean, "max_abs": _max_abs}
            self.aggregate_fn = agg_fns[aggregate_method]
        if self.aggregate_fn is not None:
            # If we aggregate over some axis, drop the corresponding axis
            self.attributions_shape = (
                orig_attributions_shape[: self.aggregate_dim]
                + orig_attributions_shape[self.aggregate_dim + 1 :]
            )
        else:
            self.attributions_shape = orig_attributions_shape
        
    def _open_attributions_file(self):
        self.attributions = h5py.File(self.path, "r")

    def __getitem__(
        self, index: int
    ) -> Tuple[int, torch.Tensor, torch.Tensor, torch.Tensor, str]:
        if self.attributions is None:
            self._open_attributions_file()
        assert self.attributions is not None
        method_idx = index // self.num_samples
        method_name = self.method_names[method_idx]
        sample_idx = index % self.num_samples
        sample, label = self.samples_dataset[sample_idx]
        dataset = _check_is_dataset_or_tensor(self.attributions[method_name])
        attrs = dataset[sample_idx]
        if not isinstance(attrs, torch.Tensor):
            attrs = torch.tensor(attrs)
        if self.aggregate_fn is not None:
            attrs = self.aggregate_fn(attrs, dim=self.aggregate_dim)
        return sample_idx, sample, label, attrs, method_name

    def __len__(self):
        return self.num_samples * len(self.method_names)


class GroupedAttributionsDataset(IndexDataset):
    def __init__(self, dataset: AttributionsDataset):
        super().__init__(dataset)
        self.dataset: AttributionsDataset = dataset
        self.method_names = dataset.method_names
        self.num_samples = dataset.num_samples
        self.attributions_shape = dataset.attributions_shape

    def __getitem__(self, index):
        if self.dataset.attributions is None:
            self.dataset._open_attributions_file()
            assert self.dataset.attributions is not None
        sample, label = self.dataset.samples_dataset[index]
        attrs: Dict[str, torch.Tensor] = {}
        for method_name in self.dataset.method_names:
            dataset = _check_is_dataset_or_tensor(
                self.dataset.attributions[method_name]
            )
            attrs[method_name] = dataset[index]
            if not isinstance(attrs[method_name], torch.Tensor):
                attrs[method_name] = torch.tensor(attrs[method_name])
        if self.dataset.aggregate_fn is not None:
            for method_name in self.dataset.method_names:
                attrs[method_name] = self.dataset.aggregate_fn(
                    attrs[method_name], dim=self.dataset.aggregate_dim
                )
        return index, sample, label, attrs

    def __len__(self):
        return self.dataset.num_samples



import torch
from typing import Callable, Dict, List
from torch import nn
from attribench.data.attributions_dataset._attributions_dataset import GroupedAttributionsDataset
from .._metric_worker import GroupedMetricWorker, WorkerConfig
from attribench.functional.metrics.infidelity._perturbation_generator import (
    PerturbationGenerator,
)
from attribench.functional.metrics.infidelity._infidelity import (
    _infidelity_batch,
)


class InfidelityWorker(GroupedMetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: Callable[[], nn.Module],
        dataset: GroupedAttributionsDataset,
        batch_size: int,
        perturbation_generators: Dict[str, PerturbationGenerator],
        num_perturbations: int,
        activation_fns: List[str],
    ):
        super().__init__(
            worker_config,
            model_factory,
            dataset,
            batch_size,
        )
        self.activation_fns = activation_fns
        self.num_perturbations = num_perturbations
        self.perturbation_generators = perturbation_generators

    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: Dict[str, torch.Tensor],
    ):
        return _infidelity_batch(
            self.model,
            batch_x,
            batch_y,
            batch_attr,
            self.perturbation_generators,
            self.num_perturbations,
            self.activation_fns,
            self.device,
        )



import pandas as pd
from typing import Dict, Tuple
from scipy import stats
import matplotlib.pyplot as plt
import seaborn as sns


class CLESPlot:
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs

    def render(self, method1, method2):
        result_cles = {}
        for key in self.dfs:
            df, higher_is_better = self.dfs[key]
            if not higher_is_better:
                df = -df

            statistic, pvalue = stats.wilcoxon(
                x=df[method1], y=df[method2], alternative="two-sided"
            )
            cles = (df[method1] > df[method2]).mean()
            result_cles[key] = cles if pvalue < 0.01 else 0.5
        sns.set_color_codes("muted")
        df = (
            pd.DataFrame(result_cles, index=["CLES"])
            .transpose()
            .reset_index()
            .rename(columns={"index": "Metric"})
        )
        df["CLES"] -= 0.5
        fig, ax = plt.subplots(figsize=(5, 7))
        sns.barplot(data=df, x="CLES", y="Metric", color="b", left=0.5, ax=ax)
        ax.set(xlim=(0, 1))
        ax.set_title(f"P({method1} > {method2})")
        return fig


import torch
from torch import nn
from typing import List, Union, Mapping, Dict
from attribench.masking.image import ImageMasker
from attribench.functional.metrics.deletion._dataset import IrofDataset
from attribench.functional.metrics.deletion._get_predictions import (
    get_predictions,
)
from attribench.data import AttributionsDataset
from torch.utils.data import DataLoader
from attribench.result._deletion_result import DeletionResult
from attribench.result._batch_result import BatchResult


def _irof_batch(
    samples: torch.Tensor,
    labels: torch.Tensor,
    model: nn.Module,
    attrs: torch.Tensor,
    maskers: Mapping[str, ImageMasker],
    activation_fns: List[str],
    mode: str,
    start: float,
    stop: float,
    num_steps: int,
) -> Dict:
    result_dict = {}
    for masker_name, masker in maskers.items():
        masking_dataset = IrofDataset(
            mode, start, stop, num_steps, samples, masker
        )
        masking_dataset.set_attrs(attrs)
        result_dict[masker_name] = get_predictions(
            masking_dataset, labels, model, activation_fns
        )
    return result_dict


def irof(
    model: nn.Module,
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    maskers: Mapping[str, ImageMasker],
    activation_fns: Union[List[str], str] = "linear",
    mode: str = "morf",
    start: float = 0.0,
    stop: float = 1.0,
    num_steps: int = 100,
    device: torch.device = torch.device("cpu"),
):
    """Computes the IROF metric for a given :class:`~attribench.data.AttributionsDataset` and model.

    IROF starts segmenting the input image using SLIC. Then, it iteratively
    masks out the top (Most Relevant First, or MoRF) or bottom (Least Relevant
    First, or LeRF) segments and computes the confidence of the model on the
    masked samples. The relevance of a segment is computed as the average
    relevance of the features in the segment.

    This results in a curve of confidence vs. number of segments masked. The
    area under (or equivalently over) this curve is the IROF metric.

    `start`, `stop`, and `num_steps` are used to determine the range of segments
    to mask. The range is determined by `start` and `stop` as a percentage of
    the total number of segments. `num_steps` is the number of steps to take
    between `start` and `stop`.

    The IROF metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`.

    Parameters
    ----------
    model : nn.Module
        Model to compute IROF on.
    attributions_dataset : AttributionsDataset
        Dataset of attributions to compute IROF on.
    batch_size : int
        Batch size to use when computing model predictions on masked samples.
    maskers : Mapping[str, ImageMasker]
        Dictionary of maskers to use for masking samples.
    activation_fns : Union[List[str], str], optional
        List of activation functions to use when computing model predictions on
        masked samples. If a single string is passed, it is converted to a
        single-element list.
        Default: "linear"
    mode : str, optional
        Mode to use when masking samples. Must be "morf" or "lerf".
        Default: "morf"
    start : float, optional
        Relative start of the range of segments to mask. Must be between 0 and 1.
        Default: 0.0
    stop : float, optional
        Relative stop of the range of segments to mask. Must be between 0 and 1.
        Default: 1.0
    num_steps : int, optional
        Number of steps to take between `start` and `stop`.
        Default: 100
    """
    if isinstance(activation_fns, str):
        activation_fns = [activation_fns]

    model.to(device)
    model.eval()

    dataloader = DataLoader(
        attributions_dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    result = DeletionResult(
        attributions_dataset.method_names,
        list(maskers.keys()),
        activation_fns,
        mode,
        num_samples=attributions_dataset.num_samples,
        num_steps=num_steps,
    )

    for (
        batch_indices,
        batch_x,
        batch_y,
        batch_attr,
        method_names,
    ) in dataloader:
        batch_x = batch_x.to(device)
        batch_y = batch_y.to(device)
        batch_result = _irof_batch(
            batch_x,
            batch_y,
            model,
            batch_attr,
            maskers,
            activation_fns,
            mode,
            start,
            stop,
            num_steps,
        )
        result.add(BatchResult(batch_indices, batch_result, method_names))



import pandas as pd
import matplotlib.pyplot as plt
from typing import Dict, Tuple
import seaborn as sns


def _create_fig(df, figsize, annot):
    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(
        df,
        annot=annot,
        vmin=-1,
        vmax=1,
        cmap=sns.diverging_palette(220, 20, as_cmap=True),
        ax=ax,
        fmt=".2f",
        cbar=False,
    )
    ax.set_aspect("equal")
    return fig, ax


class InterMetricCorrelationPlot:
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs
        self.methods = list(dfs.values())[0][0].columns

    def render(self, figsize, annot, fontsize):
        corr_dfs = []
        for method_name in self.methods:
            data = {}
            for metric_name, (df, higher_is_better) in self.dfs.items():
                data[metric_name] = (
                    -df[method_name].to_numpy()
                    if not higher_is_better
                    else df[method_name].to_numpy()
                )
            df = pd.DataFrame(data)
            corr_dfs.append(df.corr(method="spearman"))
        corr = pd.concat(corr_dfs).groupby(level=0).mean()
        corr = corr.reindex(corr.columns)
        fig, ax = _create_fig(corr, figsize, annot)

        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            ha="right",
            rotation_mode="anchor",
            fontsize=fontsize,
        )
        ax.set_yticklabels(ax.get_yticklabels(), fontsize=fontsize)
        return fig
        


class InterMethodCorrelationPlot:
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs

    def render(self, figsize=(20, 20), annot=True):
        # Compute correlations for each metric
        all_dfs = [
            df if not inverted else -df
            for _, (df, inverted) in self.dfs.items()
        ]
        corr_dfs = [df.corr(method="spearman") for df in all_dfs]

        # Compute average of correlations
        corr = pd.concat(corr_dfs).mean(level=0)
        return _create_fig(corr, figsize, annot)

    def render_all(self, figsize=(20, 20), annot=True):
        figs = {}
        for name, (df, inverted) in self.dfs.items():
            if inverted:
                df = -df
            corr = df.corr(method="spearman")
            figs[name] = _create_fig(corr, figsize, annot)
        return figs



from attribench.masking import ImageMasker
from cv2 import blur
import numpy as np
import torch


class BlurringMasker(ImageMasker):
    def __init__(self, feature_level, kernel_size):
        super().__init__(feature_level)
        if not 0 < kernel_size < 1.0:
            raise ValueError(
                "Kernel size is expressed as a fraction of image height, and must be between 0 and 1."
            )
        self.kernel_size = kernel_size

    def _initialize_baselines(self, samples: torch.tensor):
        kernel_size = int(self.kernel_size * samples.shape[-1])

        baseline = []
        for i in range(samples.shape[0]):
            sample = samples[i, ...].cpu().numpy()
            cv_sample = np.transpose(sample, (1, 2, 0))
            blurred_sample = blur(cv_sample, (kernel_size, kernel_size))
            if len(blurred_sample.shape) == 2:
                blurred_sample = blurred_sample[..., np.newaxis]
            baseline.append(np.transpose(blurred_sample, (2, 0, 1)))
        self.baseline = torch.tensor(
            np.stack(baseline, axis=0), device=samples.device
        )



from .hdf5_dataset._hdf5_dataset import HDF5Dataset
from .hdf5_dataset._hdf5_dataset_writer import HDF5DatasetWriter
from ._index_dataset import IndexDataset
from .attributions_dataset._attributions_dataset import AttributionsDataset
from .attributions_dataset._attributions_dataset_writer import AttributionsDatasetWriter



from attribench.masking import Masker
from attribench._segmentation import segment_attributions
from typing import List, Union, Optional, Tuple
import numpy as np
import torch
from abc import abstractmethod


class ImageMasker(Masker):
    """Abstract base class for all image maskers.

    Image maskers are specific maskers for image data. They can be used to
    mask features on the feature level or the pixel level. If the masker is
    used on the feature level, the attributions must have the same shape as
    the samples. If the masker is used on the pixel level, the attributions
    must have the same shape as the samples, except the channel dimension
    must be 1.

    If the masker is used on the feature level, masking a feature means
    masking a specific input feature for all samples (i.e. one color value for
    one pixel). If the masker is used on the pixel level, masking a feature
    means masking a specific pixel for all samples (i.e. all color values for
    one pixel).

    If segmented samples are provided, the masker will use the segments to
    mask features. This means that masking a feature will mask all pixels
    belonging to the same segment. The attribution value of a segment is
    defined as the average attribution value of all features in that segment.
    """

    def __init__(self, masking_level: str):
        """
        Parameters
        ----------
        masking_level : str
            Either ``"feature"`` or ``"pixel"``. If ``"feature"``, the masker
            will mask features. If ``"pixel"``, the masker will mask pixels.

        Raises
        ------
        ValueError
            If ``masking_level`` is not ``"feature"`` or ``"pixel"``.
        """
        if masking_level not in ("feature", "pixel"):
            raise ValueError(
                f"feature_level must be 'feature' or 'pixel'."
                f" Found {masking_level}."
            )
        self.masking_level = masking_level
        super().__init__()
        # will be set after initialize_batch:
        self.segmented_samples: Optional[torch.Tensor] = None
        self.segmented_attributions: Optional[np.ndarray] = None
        self.segment_indices: Optional[List[np.ndarray]] = None
        self.use_segments: bool = False
        self.sorted_indices: torch.Tensor | List[torch.Tensor] | None = None

    def set_batch(
        self,
        samples: torch.Tensor,
        attributions: torch.Tensor | None = None,
        segmented_samples: torch.Tensor | None = None,
    ):
        """Set the batch of samples and attributions to use for masking.
        Optionally also set the segmented samples.

        Parameters
        ----------
        samples : torch.Tensor
            Samples to use for masking.
        attributions : torch.Tensor, optional
            Attributions to use for masking, by default None
            If None, the :meth:`mask_top` and :meth:`mask_bot` methods will
            not be available.
        segmented_samples : torch.Tensor, optional
            Segmented samples to use for masking, by default None
        """

        # Check if attributions are compatible with samples
        if attributions is not None and not self._check_attribution_shape(
            samples, attributions
        ):
            raise ValueError(
                f"samples and attribution shape not compatible for masking "
                f"level {self.masking_level}."
                f"Found shapes {samples.shape} and {attributions.shape}"
            )

        # Check if segmented samples are compatible with samples
        if segmented_samples is not None:
            if not (
                samples.shape[0] == segmented_samples.shape[0]
                and samples.shape[-2:] == segmented_samples.shape[-2:]
            ):
                raise ValueError(
                    f"Incompatible shapes: {samples.shape}, {segmented_samples.shape}"
                )
            if samples.device != segmented_samples.device:
                raise ValueError(
                    "Device for samples and segmented_samples must be equal."
                    f"Got {samples.device} for samples,"
                    f" {segmented_samples.device} for segmented_samples."
                )

        # Set attributes
        self.samples = samples
        self.attributions = attributions
        self.segmented_samples = segmented_samples
        self.use_segments = segmented_samples is not None

        if segmented_samples is not None and attributions is not None:
            # If segmented samples and attributions are given,
            # segment the attributions accordingly
            # and sort the segments
            self.segmented_attributions = segment_attributions(
                segmented_samples.cpu().numpy(),
                attributions.cpu().numpy(),
            )
            sorted_indices = torch.tensor(
                self.segmented_attributions.argsort()
            )

            # Filter out the -np.inf values from the sorted indices
            filtered_sorted_indices = []
            for i in range(segmented_samples.shape[0]):
                num_infs = np.count_nonzero(
                    self.segmented_attributions[i, ...] == -np.inf
                )
                filtered_sorted_indices.append(sorted_indices[i, num_infs:])
            self.sorted_indices = filtered_sorted_indices
        elif attributions is not None:
            # If only attributions are given, sort them
            self.sorted_indices = torch.tensor(
                attributions.cpu()
                .numpy()
                .reshape(attributions.shape[0], -1)
                .argsort()
            )

        if segmented_samples is not None:
            # Get the indices of the segments for each image
            self.segment_indices = [
                np.unique(segmented_samples.cpu().numpy()[i, ...])
                for i in range(samples.shape[0])
            ]

        self._initialize_baselines(self.samples)

    def get_num_features(self):
        assert self.samples is not None
        if self.use_segments:
            raise ValueError(
                "When using segments, total number of features varies per image."
            )
        if self.masking_level == "feature":
            return self.samples.flatten(1).shape[-1]
        if self.masking_level == "pixel":
            return self.samples.flatten(2).shape[-1]

    def mask_top(self, k: int):
        assert self.samples is not None
        assert self.sorted_indices is not None
        if not self.use_segments:
            return super().mask_top(k)
        if k == 0:
            return self.samples
        # When using segments, k is relative (between 0 and 1)
        indices = []
        for i in range(self.samples.shape[0]):
            num_segments = len(self.sorted_indices[i])
            num_to_mask = int(num_segments * k)
            indices.append(self.sorted_indices[i][-num_to_mask:])
        return self._mask_segments(indices)

    def mask_bot(self, k: int):
        assert self.samples is not None
        assert self.sorted_indices is not None
        if not self.use_segments:
            return super().mask_bot(k)
        if k == 0:
            return self.samples
        # When using segments, k is relative (between 0 and 1)
        indices = []
        for i in range(self.samples.shape[0]):
            num_segments = len(self.sorted_indices[i])
            num_to_mask = int(num_segments * k)
            indices.append(self.sorted_indices[i][:num_to_mask])
        return self._mask_segments(indices)

    def mask_rand(
        self, k: int, return_indices=False
    ) -> torch.Tensor | Tuple[torch.Tensor, torch.Tensor]:
        assert self.samples is not None
        if k == 0:
            return self.samples
        rng = np.random.default_rng()
        if not self.use_segments:
            return super().mask_rand(k, return_indices)
        else:
            # this is done a little different form super(),
            # no shuffle here: for each image, only select segments that exist in that image
            # k can be large -> rng.choice raises exception if k > number of segments
            assert self.segment_indices is not None
            assert self.segmented_samples is not None
            indices = torch.stack(
                [
                    torch.tensor(
                        rng.choice(
                            self.segment_indices[i], size=k, replace=False
                        )
                    )
                    for i in range(self.segmented_samples.shape[0])
                ]
            )
            masked_samples = self._mask_segments(indices)
        if return_indices:
            return masked_samples, indices
        return masked_samples

    def _check_attribution_shape(
        self, samples: torch.Tensor, attributions: torch.Tensor
    ):
        if self.masking_level == "feature":
            # Attributions should be same shape as samples
            return list(samples.shape) == list(attributions.shape)
        elif self.masking_level == "pixel":
            # attributions should have the same shape as samples,
            # except the channel dimension must be 1
            aggregated_shape = list(samples.shape)
            aggregated_shape[1] = 1
            return aggregated_shape == list(attributions.shape)

    def _mask(self, indices: torch.Tensor) -> torch.Tensor:
        if self.baseline is None:
            raise ValueError("Masker was not initialized.")
        if self.use_segments:
            return self._mask_segments(indices)
        else:
            assert self.samples is not None
            batch_size, _, _, _ = self.samples.shape
            num_indices = indices.shape[1]
            batch_dim = np.tile(
                range(batch_size), (num_indices, 1)
            ).transpose()

            to_mask = torch.zeros(
                self.samples.shape, device=self.samples.device
            ).flatten(1 if self.masking_level == "feature" else 2)
            # to_mask = np.zeros(self.samples.shape)
            if self.masking_level == "feature":
                to_mask[batch_dim, indices] = 1.0
            else:
                try:
                    to_mask[batch_dim, :, indices] = 1.0
                except IndexError:
                    raise ValueError(
                        "Masking index was out of bounds. "
                        "Make sure the masking policy is compatible with method output."
                    )
            return self._mask_boolean(to_mask.view(self.samples.shape).bool())

    def _mask_segments(
        self, segments: Union[torch.Tensor, List[torch.Tensor]]
    ) -> torch.Tensor:
        assert self.segmented_samples is not None
        assert self.samples is not None
        if not self.segmented_samples.shape[0] == len(segments):
            raise ValueError(
                f"Number of segment lists doesn't match number of images:"
                f" {self.segmented_samples.shape[0]}"
                f"were expected, {len(segments)} were given."
            )
        bool_masks = []
        for i in range(self.samples.shape[0]):
            seg_img = self.segmented_samples[i, ...]
            segs = segments[i].to(seg_img.device)
            bool_masks.append(_isin(seg_img, segs))
        bool_masks = torch.stack(bool_masks, dim=0)
        if self.samples.shape[1] == 3:
            bool_masks = bool_masks.repeat(1, 3, 1, 1)
        return self._mask_boolean(bool_masks)

    def _mask_boolean(self, bool_mask: torch.Tensor) -> torch.Tensor:
        return (
            self.samples
            - (bool_mask * self.samples)
            + (bool_mask * self.baseline)
        )

    @abstractmethod
    def _initialize_baselines(self, samples: torch.Tensor):
        raise NotImplementedError


def _isin(a: torch.Tensor, b: torch.Tensor):
    # https://stackoverflow.com/questions/60918304/get-indices-of-elements-in-tensor-a-that-are-present-in-tensor-b
    return (a[..., None] == b).any(-1)



import numpy as np
import h5py
from typing import List, Tuple, Dict, Callable, Optional


class NDArrayTree:
    def __init__(self, levels: List[Tuple[str, List[str]]]):
        self.levels = levels

        def _initialize_data(data=None, depth=0):
            if data is None:
                data = {}
            _, keys = levels[depth]
            for key in keys:
                if depth < len(levels) - 1:
                    data[key] = {}
                    _initialize_data(data[key], depth + 1)
                else:
                    data[key] = None
            return data

        self.data = _initialize_data()

    def append(self, new_data: Dict, axis=0, **kwargs):
        def _append_rec(_data, _new_data, depth=0):
            level_name = self.levels[depth][0]
            if level_name in kwargs.keys():
                # If level name found in kwargs, descend down the corresponding branch
                # In this case we pass _new_data instead of _new_data[key], since this level is not present
                # in the data dict
                key = kwargs[level_name]
                if type(_data[key]) == dict:
                    _append_rec(_data[key], _new_data, depth + 1)
                elif type(_data[key]) == np.ndarray or _data[key] is None:
                    _data[key] = (
                        _new_data
                        if _data[key] is None
                        else np.concatenate([_data[key], _new_data], axis=axis)
                    )
                else:
                    raise ValueError(f"Invalid type: {type(_data[key])}")
            else:
                # If level name not found in kwargs, loop over each key in current level
                for key in _new_data:
                    if type(_new_data[key]) in (dict, h5py.Group) or type(
                        _data[key]
                    ) in (dict, h5py.Group):
                        # Descend down the tree (passing down a level both in _data and in _new_data)
                        _append_rec(_data[key], _new_data[key], depth + 1)
                    elif type(_new_data[key]) == np.ndarray:
                        # Base case: leaf nodes are ndarrays
                        _data[key] = (
                            _new_data[key]
                            if _data[key] is None
                            else np.concatenate(
                                [_data[key], _new_data[key]], axis=axis
                            )
                        )
                    elif type(_new_data[key]) == h5py.Dataset:
                        # Base case: leaf nodes are HDF5 datasets, convert to ndarray using [()]
                        _data[key] = (
                            _new_data[key][()]
                            if _data[key] is None
                            else np.concatenate(
                                [_data[key], _new_data[key][()]], axis=axis
                            )
                        )
                    else:
                        raise ValueError(
                            f"Invalid type: {type(_new_data[key])}"
                        )

        _append_rec(self.data, new_data)

    def apply(self, fn: Callable):
        def _apply_rec(_cur_data):
            for key in _cur_data:
                if type(_cur_data[key]) == dict:
                    # Descend down the tree
                    _apply_rec(_cur_data[key])
                elif type(_cur_data[key]) == np.ndarray:
                    _cur_data[key] = fn(_cur_data[key])

        _apply_rec(self.data)

    def get(
        self,
        postproc_fn: Optional[Callable] = None,
        select: Optional[Dict[str, list]] = None,
        exclude: Optional[Dict[str, list]] = None,
    ):
        # Initialize empty dicts if select/exclude not given
        select = select if select is not None else dict()
        exclude = exclude if exclude is not None else dict()

        postproc_fn_inst = (
            postproc_fn if postproc_fn is not None else lambda x: x
        )

        def _get_rec(cur_data=None, depth=0):
            # Initialize cur_data if necessary (root level)
            if cur_data is None:
                cur_data = self.data
            # Get the level name and keys
            level, keys = self.levels[depth]
            select_keys = select.get(level, list(cur_data.keys()))
            exclude_keys = exclude.get(level, [])
            if depth == len(self.levels) - 1:
                # We are at leaf node, return desired array(s)
                return {
                    key: postproc_fn_inst(cur_data[key])
                    for key in select_keys
                    if key not in exclude_keys
                }
            else:
                # We are not at leaf node, execute recursive call
                return {
                    key: _get_rec(cur_data[key], depth + 1)
                    for key in select_keys
                    if key not in exclude_keys
                }

        return _get_rec()

    def add_to_hdf(self, group: h5py.Group):
        def _add_rec(cur_data, cur_group):
            for key in cur_data:
                if type(cur_data[key]) == np.ndarray:
                    cur_group.create_dataset(key, data=cur_data[key])
                elif type(cur_data[key]) == dict:
                    next_group = cur_group.create_group(key)
                    _add_rec(cur_data[key], next_group)

        _add_rec(self.data, group)

    @classmethod
    def load_from_hdf(cls, level_names: List[str], group: h5py.Group):
        def _load_levels(cur_group, depth=0, cur_result=None):
            if cur_result is None:
                cur_result = []
            keys = list(cur_group.keys())
            cur_result.append((level_names[depth], keys))
            if depth < len(level_names) - 1:
                return _load_levels(cur_group[keys[0]], depth + 1, cur_result)
            return cur_result

        levels = _load_levels(group)
        result = cls(levels)
        result.append(dict(group))
        return result



from typing import Union, Tuple, List, Optional, Mapping
from attribench.masking.image import ImageMasker
from attribench._model_factory import ModelFactory
from ..._worker import Worker, WorkerConfig
from ._irof_worker import IrofWorker
from ..deletion._deletion import Deletion
from attribench.data import AttributionsDataset


class Irof(Deletion):
    """Compute the IROF metric for a given :class:`~attribench.data.AttributionsDataset` and model
    using multiple processes.

    IROF starts segmenting the input image using SLIC. Then, it iteratively
    masks out the top (Most Relevant First, or MoRF) or bottom (Least Relevant
    First, or LeRF) segments and computes the confidence of the model on the
    masked samples. The relevance of a segment is computed as the average
    relevance of the features in the segment.

    This results in a curve of confidence vs. number of segments masked. The
    area under (or equivalently over) this curve is the IROF metric.

    `start`, `stop`, and `num_steps` are used to determine the range of segments
    to mask. The range is determined by `start` and `stop` as a percentage of
    the total number of segments. `num_steps` is the number of steps to take
    between `start` and `stop`.

    The IROF metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`. The number of processes is
    determined by the number of devices. If `devices` is None, then all
    available devices are used. Samples are distributed evenly across the
    processes.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        maskers: Mapping[str, ImageMasker],
        activation_fns: Union[List[str], str],
        mode: str = "morf",
        start: float = 0.0,
        stop: float = 1.0,
        num_steps: int = 100,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        attributions_dataset : AttributionsDataset
            Dataset containing the samples and attributions to compute
            IROF on.
        batch_size : int
            The batch size to use when computing the metric.
        maskers : Dict[str, Masker]
            Dictionary of maskers to use for computing the metric.
        activation_fns : Union[List[str], str]
            List of activation functions to use for computing the metric.
            If a single string is given, it is converted to a single-element
            list.
        mode : str, optional
            Mode to use for computing the metric. Either "morf" or "lerf".
            Default: "morf"
        start : float, optional
            Relative start of the range of features to mask. Must be between 0 and 1.
            Default: 0.0
        stop : float, optional
            Relative end of the range of features to mask. Must be between 0 and 1.
            Default: 1.0
        num_steps : int, optional
            Number of steps to use for the range of features to mask.
            Default: 100
        address : str, optional
            Address to use for the multiprocessing connection.
            Default: "localhost"
        port : str, optional
            Port to use for the multiprocessing connection.
            Default: "12355"
        devices : Optional[Tuple], optional
            Devices to use. If None, then all available devices are used.
            Default: None
        """
        super().__init__(
            model_factory,
            attributions_dataset,
            batch_size,
            maskers,
            activation_fns,
            mode,
            start,
            stop,
            num_steps,
            address,
            port,
            devices,
        )
        self.maskers = maskers

    def _create_worker(self, worker_config: WorkerConfig) -> Worker:
        return IrofWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.maskers,
            self.activation_fns,
            self.mode,
            self._start,
            self.stop,
            self.num_steps,
        )



from typing import Generic, TypeVar

DT = TypeVar("DT")


class Message:
    pass


class PartialResultMessage(Message, Generic[DT]):
    def __init__(self, rank: int, data: DT):
        self.rank = rank
        self.data = data


class DoneMessage(Message):
    def __init__(self, rank: int):
        self.rank = rank


from typing import Tuple, Optional, List
from typing_extensions import override
import numpy as np
import pandas as pd
from numpy import typing as npt

from ._grouped_metric_result import GroupedMetricResult


def _column_avg(
    x: npt.NDArray, columns: Optional[npt.NDArray] = None
) -> npt.NDArray:
    """
    Returns the average value of x along the specified columns.
    Used in get_df.
    """
    if columns is not None:
        x = x[..., columns]
    return np.mean(x, axis=-1)


# TODO loads of code duplication here with DeletionResult
class SensitivityNResult(GroupedMetricResult):
    """Represents results from running the Sensitivity-N metric.
    """
    def __init__(
        self,
        method_names: List[str],
        maskers: List[str],
        activation_fns: List[str],
        num_samples: int,
        num_steps: int
    ):
        """
        Parameters
        ----------
        method_names : List[str]
            Names of attribution methods tested by Sensitivity-N.
        maskers : List[str]
            Names of maskers used by Sensitivity-N.
        activation_fns : List[str]
            Names of activation functions used by Sensitivity-N.
        num_samples : int
            Number of samples on which Sensitivity-N was run.
        num_steps : int
            Number of steps used by Sensitivity-N.
        """
        levels = {
            "masker": maskers,
            "activation_fn": activation_fns,
            "method": method_names,
        }
        shape = [num_samples, num_steps]
        level_order = ["masker", "activation_fn", "method"]
        super().__init__(method_names, shape, levels, level_order)

    @classmethod
    @override
    def _load(cls, path: str, format="hdf5") -> "SensitivityNResult":
        tree = cls._load_tree(path, format)
        res = SensitivityNResult(
            tree.levels["method"],
            tree.levels["masker"],
            tree.levels["activation_fn"],
            tree.shape[0],
            tree.shape[1]
        )
        res.tree = tree
        return res

    @override
    def get_df(
        self,
        masker: str,
        activation_fn: str,
        methods: Optional[List[str]] = None,
        columns: Optional[npt.NDArray] = None,
    ) -> Tuple[pd.DataFrame, bool]:
        """
        Retrieves a dataframe from the result for a given masker and
        activation function. The dataframe contains a row for each sample and a
        column for each method. Each value is the average Sensitivity-N value
        for the given method on the given sample,
        over the specified columns.

        Parameters
        ----------
        masker : str
            the masker to use
        activation_fn : str
            the activation function to use
        methods : Optional[List[str, ...]]
            the methods to include. If None, includes all methods.
        columns : Optional[npt.NDArray]
            the columns used in the aggregation.
            If None, uses all columns.

        Returns
        -------
        Tuple[pd.DataFrame, bool]
            dataframe containing results,
            and boolean indicating if higher is better
            (always True for Sensitivity-N)
        """

        methods = methods if methods is not None else self.method_names
        df_dict = {}
        for method in methods:
            array = self.tree.get(
                masker=masker, activation_fn=activation_fn, method=method
            )
            df_dict[method] = _column_avg(array, columns)
        return pd.DataFrame.from_dict(df_dict), True



from typing import Tuple
from .._typing import _check_is_dataset
import h5py
from numpy import typing as npt


class AttributionsDatasetWriter:
    def __init__(self, path: str, num_samples: int):
        self.path = path
        self.num_samples: int = num_samples
        self.sample_shape: Tuple[int, ...] | None = None
        self.file = h5py.File(self.path, "w")
        self.file.attrs["num_samples"] = self.num_samples

    def write(
        self, indices: npt.NDArray, attributions: npt.NDArray, method_name: str
    ):
        if self.sample_shape is None:
            self.sample_shape = attributions.shape[1:]
            self.file.attrs["sample_shape"] = self.sample_shape
        if attributions.shape[1:] != self.sample_shape:
            raise ValueError(
                f"Invalid sample shape. Expected: {self.sample_shape}, "
                f"got: {attributions.shape[1:]}"
            )
        if method_name not in self.file.keys():
            dataset = self.file.create_dataset(
                method_name, shape=(self.num_samples, *self.sample_shape)
            )
        else:
            dataset = _check_is_dataset(self.file[method_name])
        dataset[indices, ...] = attributions



import numpy as np
from numpy import typing as npt
import torch

from attribench.masking import Masker
from attribench.masking.image import ImageMasker
from attribench._segmentation import segment_samples


class SensitivityNDataset:
    def __init__(
        self,
        n_range: npt.NDArray[np.int32],
        num_subsets: int,
        samples: torch.Tensor,
        masker: Masker,
    ):
        self.n_range = n_range
        self.num_subsets = num_subsets
        self.samples = samples
        self.masker = masker
        self.masker.set_batch(samples)
        self.rng = np.random.default_rng()

    def __len__(self):
        return self.n_range.shape[0] * self.num_subsets

    def __getitem__(self, item):
        # This is implicitly a nested loop, running through num_subsets and
        # n_range. The inner loop is the num_subsets loop, the outer loop is
        # the n_range loop.
        n_idx = item // self.num_subsets
        subset_idx = item % self.num_subsets
        masked_samples, indices = self.masker.mask_rand(
            int(self.n_range[n_idx]), return_indices=True
        )
        return masked_samples, indices, self.n_range[n_idx], subset_idx


class SegSensNDataset:
    def __init__(
        self,
        n_range: npt.NDArray[np.int32],
        num_subsets: int,
        samples: torch.Tensor,
    ):
        self.n_range = n_range
        self.num_subsets = num_subsets
        self.samples = samples
        self.segmented_images = torch.tensor(
            segment_samples(samples.cpu().numpy()), device=self.samples.device
        )
        self.masker: ImageMasker | None = None

    def __len__(self):
        return self.n_range.shape[0] * self.num_subsets

    def __getitem__(self, item):
        assert self.masker is not None
        n_idx = item // self.num_subsets
        subset_idx = item % self.num_subsets
        masked_samples, indices = self.masker.mask_rand(
            int(self.n_range[n_idx]), return_indices=True
        )
        return masked_samples, indices, self.n_range[n_idx], subset_idx

    def set_masker(self, masker: ImageMasker):
        self.masker = masker
        self.masker.set_batch(
            self.samples, segmented_samples=self.segmented_images
        )



import numpy as np
import torch
from typing import Optional


class Masker:
    def __init__(self):
        self.baseline: Optional[torch.Tensor] = None
        self.samples: Optional[torch.Tensor] = None
        self.attributions: Optional[np.ndarray] = None
        self.sorted_indices: Optional[np.ndarray] = None
        self.rng = np.random.default_rng()

    def set_batch(self, samples: torch.tensor, attributions: np.ndarray = None):
        raise NotImplementedError

    def get_num_features(self):
        return self.sorted_indices.shape[1]

    def mask_top(self, k):
        if k == 0:
            return self.samples
        else:
            return self._mask(self.sorted_indices[:, -k:])

    def mask_bot(self, k):
        return self._mask(self.sorted_indices[:, :k])

    def mask_rand(self, k, return_indices=False):
        if k == 0:
            return self.samples

        num_samples = self.samples.shape[0]
        num_features = self.get_num_features()

        indices = np.tile(self.rng.choice(num_features, size=k, replace=False), (num_samples, 1))
        masked_samples = self._mask(indices)
        if return_indices:
            return masked_samples, indices
        return masked_samples

    def _check_attribution_shape(self, samples, attributions):
        raise NotImplementedError

    def _mask(self, indices: np.ndarray):
        raise NotImplementedError

    def _mask_boolean(self, bool_mask):
        raise NotImplementedError




import pandas as pd
from krippendorff import krippendorff
from scipy.stats import rankdata
from typing import Dict, Tuple
import matplotlib.pyplot as plt


class KrippendorffAlphaPlot:
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs

    def render(self, title=None, fontsize=20, figsize=(10, 10)):
        k_a = {
            metric_name: krippendorff.alpha(
                rankdata(df.to_numpy(), axis=1), level_of_measurement="ordinal"
            )
            for metric_name, (df, _) in self.dfs.items()
        }
        k_a = pd.DataFrame(k_a, index=["Krippendorff Alpha"]).transpose()
        fig, ax = plt.subplots()
        plt.xticks(fontsize=fontsize)
        plt.yticks(fontsize=fontsize)
        k_a.plot.barh(figsize=figsize, ax=ax)
        ax.set_title(title)
        fig.tight_layout()
        return fig



from typing import List, Mapping, Optional, Union
from attribench.masking import Masker
from attribench.data import AttributionsDataset
from torch import nn
import torch
from attribench.functional.metrics.deletion._deletion import _deletion_batch
from attribench.result import InsertionResult
from attribench.result._batch_result import BatchResult
from torch.utils.data import DataLoader


def insertion(
    model: nn.Module,
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    maskers: Mapping[str, Masker],
    activation_fns: Union[List[str], str] = "linear",
    mode: str = "morf",
    start: float = 0.0,
    stop: float = 1.0,
    num_steps: int = 100,
    device: Optional[torch.device] = None,
):
    """Computes the Insertion metric for a given :class:`~attribench.data.AttributionsDataset` and model.
    Insertion can be viewed as an opposite version of the Deletion metric.

    Insertion is computed by iteratively revealing the top (Most Relevant First,
    or MoRF) or bottom (Least Relevant First, or LeRF) features of
    the input samples, leaving the other features masked out,
    and computing the confidence of the model on the masked
    samples.

    This results in a curve of confidence vs. number of features masked. The
    area under (or equivalently over) this curve is the Insertion metric.

    `start`, `stop`, and `num_steps` are used to determine the range of features
    to mask. The range is determined by `start` and `stop` as a percentage of
    the total number of features. `num_steps` is the number of steps to take
    between `start` and `stop`.

    The Insertion metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`.

    Parameters
    ----------
    model : nn.Module
        Model to compute Insertion on.
    attributions_dataset : AttributionsDataset
        Dataset of attributions to compute Insertion on.
    batch_size : int
        Batch size to use when computing model predictions on masked samples.
    maskers : Dict[str, Masker]
        Dictionary of maskers to use for masking samples.
    activation_fns : Union[List[str], str], optional
        List of activation functions to use when computing model predictions on
        masked samples. If a single string is passed, it is converted to a
        single-element list.
        Default: "linear"
    mode : str, optional
        Mode to use when masking samples. Must be "morf" or "lerf".
        Default: "morf"
    start : float, optional
        Relative start of the range of features to mask. Must be between 0 and 1.
        Default: 0.0
    stop : float, optional
        Relative end of the range of features to mask. Must be between 0 and 1.
        Default: 1.0
    num_steps : int, optional
        Number of steps to use for the range of features to mask.
        Default: 100
    device : Optional[torch.device], optional
        Device to use, by default `None`.
        If `None`, the CPU is used.

    Returns
    -------
    InsertionResult
        Result of the Insertion metric.
    """
    if device is None:
        device = torch.device("cpu")
    if isinstance(activation_fns, str):
        activation_fns = [activation_fns]

    dataloader = DataLoader(
        attributions_dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    result = InsertionResult(
        attributions_dataset.method_names,
        list(maskers.keys()),
        activation_fns,
        mode,
        attributions_dataset.num_samples,
        num_steps,
    )

    for (
        batch_indices,
        batch_x,
        batch_y,
        batch_attr,
        method_names,
    ) in dataloader:
        batch_x = batch_x.to(device)
        batch_y = batch_y.to(device)
        batch_result = _deletion_batch(
            batch_x,
            batch_y,
            model,
            batch_attr.numpy(),
            maskers,
            activation_fns,
            "morf" if mode == "lerf" else "lerf",  # swap mode
            1 - start,  # swap start
            1 - stop,  # swap stop
            num_steps,
        )
        result.add(BatchResult(batch_indices, batch_result, method_names))
    return result



from .._metric import Metric
from typing import Dict, Tuple, Optional, List
from attribench.data.attributions_dataset._attributions_dataset import GroupedAttributionsDataset, AttributionsDataset
from ._infidelity_worker import InfidelityWorker
from ..._worker import WorkerConfig
from attribench.result import InfidelityResult
from attribench.functional.metrics.infidelity._perturbation_generator import (
    PerturbationGenerator,
)
from attribench._model_factory import ModelFactory


class Infidelity(Metric):
    """Computes the Infidelity metric for a given :class:`~attribench.data.AttributionsDataset` and
    model using multiple processes.

    Infidelity is computed by generating perturbations for each sample in the
    dataset and computing the difference in the model's output on the original
    sample and the perturbed sample. This difference is then compared to the
    dot product of the perturbation vector and the attribution map for each
    attribution method. The Infidelity metric is the mean squared error between
    these two values.

    The idea is that if the dot product is large, then the perturbation vector
    is aligned with the attribution map, and the model's output should change
    significantly when the perturbation is applied. If the dot product is small,
    then the perturbation vector is not aligned with the attribution map, and
    the model's output should not change significantly when the perturbation is
    applied.

    The mean squared error is computed for `num_perturbations` perturbations
    for each sample. The `perturbation_generators` argument is a dictionary
    mapping perturbation generator names to `PerturbationGenerator` objects.
    These objects can be used to implement different versions of Infidelity.

    The Infidelity metric is computed for each perturbation generator in
    `perturbation_generators` and each activation function in `activation_fns`.
    The number of processes is determined by the number of devices. If `devices`
    is None, then all available devices are used. Samples are distributed evenly
    across the processes.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        activation_fns: List[str],
        perturbation_generators: Dict[str, PerturbationGenerator],
        num_perturbations: int,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        attributions_dataset : AttributionsDataset
            Dataset containing the samples and attributions to compute
            Infidelity on.
        batch_size : int
            Batch size to use when computing Infidelity.
        activation_fns : Tuple[str]
            Tuple of activation functions to use when computing Infidelity.
        perturbation_generators : Dict[str, PerturbationGenerator]
            Dictionary of perturbation generators to use for generating
            perturbations.
        num_perturbations : int
            Number of perturbations to generate for each sample.
        address : str, optional
            Address to use for the multiprocessing connection,
            by default "localhost"
        port : str, optional
            Port to use for the multiprocessing connection,
            by default "12355"
        devices : Optional[Tuple], optional
            Devices to use. If None, then all available devices are used.
            By default None.
        """
        super().__init__(
            model_factory, attributions_dataset, batch_size, address, port, devices
        )
        self.dataset = GroupedAttributionsDataset(attributions_dataset)
        self.activation_fns = activation_fns
        self.num_perturbations = num_perturbations
        self.perturbation_generators = perturbation_generators
        self._result = InfidelityResult(
            self.dataset.method_names,
            list(self.perturbation_generators.keys()),
            self.activation_fns,
            num_samples=self.dataset.num_samples,
        )

    def _create_worker(self, worker_config: WorkerConfig) -> InfidelityWorker:
        return InfidelityWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.perturbation_generators,
            self.num_perturbations,
            self.activation_fns,
        )



from attribench.data.attributions_dataset._attributions_dataset import (
    AttributionsDataset,
    GroupedAttributionsDataset
)
from .._metric_worker import MetricWorker
from ..._worker import WorkerConfig
from .._metric import Metric
from typing import Optional, Tuple
from torch.utils.data import Dataset
from attribench import MethodFactory
from attribench.result._max_sensitivity_result import MaxSensitivityResult
from ._max_sensitivity_worker import MaxSensitivityWorker
from attribench._model_factory import ModelFactory


class MaxSensitivity(Metric):
    """Compute the Max-Sensitivity metric for a given `Dataset` and attribution
    methods using multiple processes.

    Max-Sensitivity is computed by adding a small amount of uniform noise
    to the input samples and computing the norm of the difference in attributions
    between the original samples and the noisy samples.
    The maximum norm of difference is then taken as the Max-Sensitivity.

    The idea is that a small amount of noise should not change the attributions
    significantly, so the norm of the difference should be small. If the norm
    is large, then the attributions are not robust to small perturbations in the
    input.

    The number of processes is determined by the number of devices. If `devices`
    is None, then all available devices are used. Samples are distributed evenly
    across the processes. Each subprocess computes the Max-Sensitivity for a
    all attribution methods on a subset of the samples.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        method_factory: MethodFactory,
        num_perturbations: int,
        radius: float,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        attributions_dataset : Dataset
            Torch Dataset to use for computing the Max-Sensitivity.
        batch_size : int
            The batch size per subprocess to use for computing the Max-Sensitivity.
        method_factory : MethodFactory
            MethodFactory instance or callable that returns a dictionary of
            attribution methods, given a model.
        num_perturbations : int
            The number of perturbations to use for computing the Max-Sensitivity.
        radius : float
            The radius of the uniform noise to add to the input samples.
        address : str, optional
            Address to use for the multiprocessing connection.
            Default: "localhost"
        port : str, optional
            Port to use for the multiprocessing connection.
            Default: "12355"
        devices : Optional[Tuple], optional
            Tuple of devices to use for multiprocessing.
            If `None`, all available devices are used.
        """
        super().__init__(
            model_factory,
            attributions_dataset,
            batch_size,
            address,
            port,
            devices,
        )
        self.method_factory = method_factory
        self.num_perturbations = num_perturbations
        self.radius = radius
        self.dataset = GroupedAttributionsDataset(attributions_dataset)
        self._result = MaxSensitivityResult(
            method_factory.get_method_names(), num_samples=attributions_dataset.num_samples
        )

    def _create_worker(self, worker_config: WorkerConfig) -> MetricWorker:
        return MaxSensitivityWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.method_factory,
            self.num_perturbations,
            self.radius,
        )



from torch.utils.data import DataLoader
from ._dataset import (
    MinimalSubsetDeletionDataset,
    MinimalSubsetInsertionDataset,
)
from attribench.masking import Masker
from typing import Callable, Dict
import torch
from torch import nn
from attribench.data import AttributionsDataset
from attribench.result import MinimalSubsetResult
from attribench.result._batch_result import BatchResult


def minimal_subset_batch(
    samples: torch.Tensor,
    model: Callable,
    attrs: torch.Tensor,
    num_steps: float,
    maskers: Dict[str, Masker],
    mode: str,
) -> Dict[str, torch.Tensor]:
    batch_result: Dict[str, torch.Tensor] = {}
    for masker_name, masker in maskers.items():
        if mode == "deletion":
            ds = MinimalSubsetDeletionDataset(
                num_steps, samples, attrs, masker
            )
            criterion_fn = lambda pred, orig: pred != orig
        elif mode == "insertion":
            ds = MinimalSubsetInsertionDataset(
                num_steps, samples, attrs, masker
            )
            criterion_fn = lambda pred, orig: pred == orig
        else:
            raise ValueError("Mode must be deletion or insertion. Got:", mode)

        # Initialize datastructures
        masker_result = torch.tensor(
            [-1 for _ in range(ds.samples.shape[0])]
        ).int()
        flipped = torch.tensor(
            [False for _ in range(ds.samples.shape[0])]
        ).bool()
        orig_predictions = torch.argmax(model(ds.samples), dim=1)

        # The MinimalSubsetDataset is an iterator that returns batches of
        # masked samples and the number of features that were masked.
        it = iter(ds)
        batch = next(it)
        while not torch.all(flipped) and batch is not None:
            masked_samples, mask_size = batch
            # Get output of model on masked samples
            with torch.no_grad():
                masked_output = model(masked_samples)
            predictions = torch.argmax(masked_output, dim=1)

            # Check which samples were flipped to either a different class
            # (deletion) or the original class (insertion)
            criterion = criterion_fn(predictions, orig_predictions)
            new_flipped = torch.logical_or(flipped, criterion.cpu())
            # Record which samples were flipped this iteration
            flipped_this_iteration = new_flipped != flipped
            masker_result[flipped_this_iteration] = mask_size
            flipped = new_flipped
            # Get next batch
            try:
                batch = next(it)
            except StopIteration:
                break

        # Set maximum value for samples that were never flipped
        num_inputs = attrs.reshape(attrs.shape[0], -1).shape[1]
        masker_result[masker_result == -1] = num_inputs
        batch_result[masker_name] = masker_result.reshape(-1, 1)
    return batch_result


def minimal_subset(
    model: nn.Module,
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    maskers: Dict[str, Masker],
    mode: str = "deletion",
    num_steps: int = 100,
    device: torch.device = torch.device("cpu"),
) -> MinimalSubsetResult:
    """Computes Minimal Subset Deletion or Insertion for a given
    :class:`~attribench.data.AttributionsDataset` and model.

    Minimal Subset Deletion or Insertion is computed by iteratively masking
    (Deletion) or revealing (Insertion) the top features of the input samples
    and computing the prediction of the model on the masked samples.

    Minimal Subset Deletion is the minimal number of features that must be
    masked to change the model's prediction from its original prediction.
    Minimal Subset Insertion is the minimal number of features that must be
    revealed to get the model's original prediction.

    The Minimal Subset metric is computed for each masker in `maskers`.

    Parameters
    ----------
    model : nn.Module
        Model to compute the Minimal Subset metric for.
    attributions_dataset : AttributionsDataset
        Dataset containing the samples and attributions to compute
        the Minimal Subset metric for.
    batch_size : int
        Batch size to use when computing the Minimal Subset metric.
    maskers : Dict[str, Masker]
        Dictionary mapping masker names to `Masker` objects.
    mode : str, optional
        "deletion" or "insertion", by default "deletion"
    num_steps : int, optional
        Number of steps to use when computing the Minimal Subset metric,
        by default 100. More steps will result in a more accurate metric,
        but will take longer to compute.
    device : torch.device, optional
        Device to use when computing the Minimal Subset metric,
        by default torch.device("cpu")

    Returns
    -------
    MinimalSubsetResult
    """
    model.to(device)
    model.eval()

    dataloader = DataLoader(
        attributions_dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    result = MinimalSubsetResult(
        attributions_dataset.method_names,
        list(maskers.keys()),
        mode,
        num_samples=len(attributions_dataset),
    )

    for (
        batch_indices,
        batch_x,
        _,
        batch_attr,
        method_names,
    ) in dataloader:
        batch_x = batch_x.to(device)
        batch_result = minimal_subset_batch(
            batch_x, model, batch_attr, num_steps, maskers, mode
        )
        result.add(BatchResult(batch_indices, batch_result, method_names))
    return result



from abc import abstractmethod
from attribench.data import IndexDataset
from tqdm import tqdm

from .._message import PartialResultMessage
from .._distributed_computation import DistributedComputation
from ._metric_worker import MetricWorker, WorkerConfig
from attribench.result._metric_result import MetricResult
from typing import Tuple, Optional
from attribench._model_factory import ModelFactory


class Metric(DistributedComputation):
    """Abstract base class for metrics that are computed using multiple processes.
    """
    def __init__(
        self,
        model_factory: ModelFactory,
        dataset: IndexDataset,
        batch_size: int,
        address: str,
        port: str | int,
        devices: Optional[Tuple] = None,
    ):
        super().__init__(address, port, devices)
        self.batch_size = batch_size
        self.dataset = dataset
        self.model_factory = model_factory
        self.prog = None  # TQDM progress bar
        self._result: Optional[MetricResult] = None

    @abstractmethod
    def _create_worker(self, worker_config: WorkerConfig) -> MetricWorker:
        raise NotImplementedError

    def _cleanup(self):
        if self.prog is not None:
            self.prog.close()

    def run(self, result_path: Optional[str] = None, progress=True):
        """
        Runs the metric computation and optionally saves the result.
        If no result path is given, the result will not be saved to disk.
        It can still be accessed via the ``result`` property.

        Parameters
        ----------
        result_path : str, optional
            Path to save the result to. If None, the result is not saved to disk.
        progress : bool, optional
            Whether to show a progress bar. Defaults to True.
        """
        if progress:
            self.prog = tqdm(total=len(self.dataset))
        super().run()
        if result_path is not None:
            self.save_result(result_path)

    def save_result(self, path: str, format="hdf5"):
        """Save the result to disk.

        Parameters
        ----------
        path : str
            Path to save the result to.
        format : str, optional
            Format to save the result in.
            If ``"hdf5"``, the result is saved as an HDF5 file.
            If ``"csv"``, the result is saved as a directory structure containing
            CSV files. Default: ``"hdf5"``.

        Raises
        ------
        ValueError
            If the result is None.
        """
        if self._result is not None:
            self._result.save(path, format)
        else:
            raise ValueError("Cannot save result: result is None")

    def _handle_result(self, result_message: PartialResultMessage):
        if self._result is not None:
            self._result.add(result_message.data)
        if self.prog is not None:
            self.prog.update(len(result_message.data.indices))

    @property
    def result(self) -> MetricResult:
        if self._result is None:
            raise ValueError("Cannot get result: result is None")
        return self._result
